from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Dict,
        Literal,
        Protocol,
        Sequence,
        TypeVar,
    )

    import numpy as np
    import numpy.typing as npt
    import pandas as pd
    from matplotlib.artist import Artist
    from matplotlib.axes import Axes
    from matplotlib.axis import Axis, XAxis, XTick, YAxis, YTick
    from matplotlib.figure import Figure
    from matplotlib.offsetbox import DrawingArea
    from matplotlib.patches import PathPatch
    from matplotlib.text import Text
    from mizani.transforms import trans
    from patsy.eval import EvalEnvironment
    from typing_extensions import TypeAlias

    from plotnine.coords.coord import coord
    from plotnine.facets.facet import facet
    from plotnine.facets.facet_grid import facet_grid
    from plotnine.facets.facet_wrap import facet_wrap
    from plotnine.facets.layout import Layout
    from plotnine.facets.strips import Strips
    from plotnine.geoms.geom import geom
    from plotnine.ggplot import ggplot
    from plotnine.guides.guide import guide
    from plotnine.iapi import strip_label_details
    from plotnine.layer import Layers, layer
    from plotnine.mapping.aes import aes
    from plotnine.positions.position import position
    from plotnine.scales.scale import scale, scale_continuous, scale_discrete
    from plotnine.scales.scales import Scales
    from plotnine.stats.stat import stat
    from plotnine.themes.theme import theme
    from plotnine.watermark import watermark

    class PlotAddable(Protocol):
        """
        Object that can be added to a ggplot object
        """

        def __radd__(self, other: ggplot) -> ggplot:
            """
            Add to ggplot object
            """
            ...

    class DataFrameConvertible(Protocol):
        """
        Object that can be converted to a DataFrame
        """

        def to_pandas(self) -> pd.DataFrame:
            """
            Convert to pandas dataframe
            """
            ...

    # Tuples
    TupleInt2: TypeAlias = tuple[int, int]
    TupleFloat2: TypeAlias = tuple[float, float]
    TupleFloat3: TypeAlias = tuple[float, float, float]
    TupleFloat4: TypeAlias = tuple[float, float, float, float]

    # Arrays (strictly numpy)
    AnyArray: TypeAlias = npt.NDArray[Any]
    BoolArray: TypeAlias = npt.NDArray[np.bool_]
    FloatArray: TypeAlias = npt.NDArray[np.float64]
    IntArray: TypeAlias = npt.NDArray[np.int64]
    StrArray: TypeAlias = npt.NDArray[np.str_]

    # Series
    AnySeries: TypeAlias = pd.Series[Any]
    IntSeries: TypeAlias = pd.Series[int]
    FloatSeries: TypeAlias = pd.Series[float]

    # ArrayLikes
    AnyArrayLike: TypeAlias = AnyArray | pd.Series[Any] | Sequence[Any]
    IntArrayLike: TypeAlias = IntArray | IntSeries | Sequence[int]
    FloatArrayLike: TypeAlias = FloatArray | FloatSeries | Sequence[float]

    # Type Variables
    # A array variable we can pass to a transforming function and expect
    # result to be of the same type
    FloatArrayLikeTV = TypeVar(
        "FloatArrayLikeTV",
        # We cannot use FloatArrayLike type because pyright expect
        # the result to be a FloatArrayLike
        FloatArray,
        FloatSeries,
        Sequence[float],
        TupleFloat2,
    )

    # Column transformation function
    TransformCol = Callable[[FloatSeries], FloatSeries | FloatArray]

    # Input data can be a DataFrame, a DataFrame factory or things that
    # are convertible to DataFrames.
    # `Data` is mostly used internally and `DataLike` is the input type
    # before automatic conversion.
    # Input data can actually also contain DataFrameGroupBy which is
    # specially handled, but pandas doesn't expose that data type in
    # their type stubs and instead treats it the same as a DataFrame
    # (df.groupby() returns a DataFrame in the stubs).
    Data: TypeAlias = pd.DataFrame | Callable[[pd.DataFrame], pd.DataFrame]
    DataLike: TypeAlias = Data | DataFrameConvertible

    LayerData: TypeAlias = (
        pd.DataFrame | Callable[[pd.DataFrame], pd.DataFrame]
    )
    LayerDataLike: TypeAlias = LayerData | DataFrameConvertible
    ColorLike: TypeAlias = str | Literal["None", "none"]
    ColorsLike: TypeAlias = (
        ColorLike | list[ColorLike] | pd.Series[ColorLike] | StrArray
    )

    # Mizani
    Trans: TypeAlias = trans

    # Facet strip
    StripLabellingFuncNames: TypeAlias = Literal[
        "label_value", "label_both", "label_context"
    ]

    # Function that can facet strips
    StripLabellingFunc: TypeAlias = Callable[
        [strip_label_details], strip_label_details
    ]

    StripLabellingDict: TypeAlias = (
        Dict[str, str] | Dict[str, Callable[[str], str]]
    )

    # Can be coerced to a StripLabellingFunc
    CanBeStripLabellingFunc: TypeAlias = (
        StripLabellingFuncNames
        | StripLabellingFunc
        | Callable[[str], str]
        | StripLabellingDict
    )

    StripPosition: TypeAlias = Literal["top", "right"]

    # Plotnine Classes
    Aes: TypeAlias = aes
    Coord: TypeAlias = coord
    Facet: TypeAlias = facet
    FacetGrid: TypeAlias = facet_grid
    FacetWrap: TypeAlias = facet_wrap
    Geom: TypeAlias = geom
    Ggplot: TypeAlias = ggplot
    Guide: TypeAlias = guide
    Layer: TypeAlias = layer
    Position: TypeAlias = position
    Scale: TypeAlias = scale
    ScaleContinuous: TypeAlias = scale_continuous
    ScaleDiscrete: TypeAlias = scale_discrete
    Stat: TypeAlias = stat
    Theme: TypeAlias = theme
    Watermark: TypeAlias = watermark

    ## Scales

    # Name names of scaled aesthetics
    ScaledAestheticsName: TypeAlias = Literal[
        "x",
        "xmin",
        "xmax",
        "xend",
        "xintercept",
        "y",
        "ymin",
        "ymax",
        "yend",
        "yintercept",
        "alpha",
        "color",
        "colour",
        "fill",
        "linetype",
        "shape",
        "size",
        "stroke",
    ]

    # limits
    ScaleContinuousLimits: TypeAlias = TupleFloat2
    ScaleDiscreteLimits: TypeAlias = Sequence[str]
    ScaleLimits: TypeAlias = ScaleContinuousLimits | ScaleDiscreteLimits

    ScaleLimitsRaw: TypeAlias = (
        None | ScaleLimits | Callable[[ScaleLimits], ScaleLimits]
    )
    ScaleContinuousLimitsRaw: TypeAlias = (
        None
        | ScaleContinuousLimits
        | Callable[[ScaleContinuousLimits], ScaleContinuousLimits]
    )
    ScaleDiscreteLimitsRaw: TypeAlias = (
        None
        | ScaleDiscreteLimits
        | Callable[[ScaleDiscreteLimits], ScaleDiscreteLimits]
    )

    # Breaks
    ScaleContinuousBreaks: TypeAlias = Sequence[float]
    ScaleDiscreteBreaks: TypeAlias = Sequence[str]
    ScaleBreaks: TypeAlias = ScaleContinuousBreaks | ScaleDiscreteBreaks

    ScaleBreaksRaw: TypeAlias = (
        bool | None | ScaleBreaks | Callable[[ScaleLimits], ScaleBreaks]
    )
    ScaleContinuousBreaksRaw: TypeAlias = (
        bool
        | None
        | ScaleContinuousBreaks
        | Callable[[ScaleContinuousLimits], ScaleContinuousBreaks]
    )
    ScaleDiscreteBreaksRaw: TypeAlias = (
        bool
        | None
        | ScaleDiscreteBreaks
        | Callable[[ScaleDiscreteLimits], ScaleDiscreteBreaks]
    )
    ScaleMinorBreaksRaw: TypeAlias = ScaleContinuousBreaksRaw | int

    # Labels
    ScaleLabelsRaw: TypeAlias = (
        bool
        | None
        | Sequence[str]
        | Callable[[ScaleBreaks], Sequence[str]]
        | dict[str, str]
    )
    ScaleLabels: TypeAlias = Sequence[str]

    ## Coords
    CoordRange: TypeAlias = TupleFloat2

    # Legend
    LegendPosition: TypeAlias = (
        Literal["left", "right", "top", "bottom"] | tuple[float, float]
    )



from ..doctools import document
from .geom_point import geom_point


@document
class geom_qq(geom_point):
    """
    Quantile-Quantile plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_PARAMS = {"stat": "qq", "position": "identity", "na_rm": False}



"""
Provide theamables, the elements of plot can be style with theme()

From the ggplot2 documentation the axis.title inherits from text.
What this means is that axis.title and text have the same elements
that may be themed, but the scope of what they apply to is different.
The scope of text covers all text in the plot, axis.title applies
only to the axis.title. In matplotlib terms this means that a theme
that covers text also has to cover axis.title.
"""
from __future__ import annotations

import typing
from contextlib import suppress
from copy import deepcopy
from typing import Dict
from warnings import warn

from ..exceptions import PlotnineError
from ..utils import RegistryHierarchyMeta, to_rgba
from .elements import element_base, element_blank

if typing.TYPE_CHECKING:
    from typing import Any, Mapping, Type

    from matplotlib.patches import Patch

    from plotnine.typing import Axes, Figure, Theme


class themeable(metaclass=RegistryHierarchyMeta):
    """
    Abstract class of things that can be themed.

    Every subclass of themeable is stored in a dict at
    :python:`themeable.register` with the name of the subclass as
    the key.

    It is the base of a class hierarchy that uses inheritance in a
    non-traditional manner. In the textbook use of class inheritance,
    superclasses are general and subclasses are specializations. In some
    since the hierarchy used here is the opposite in that superclasses
    are more specific than subclasses.

    It is probably better to think if this hierarchy of leveraging
    Python's multiple inheritance to implement composition. For example
    the ``axis_title`` themeable is *composed of* the ``x_axis_title`` and the
    ``y_axis_title``. We are just using multiple inheritance to specify
    this composition.

    When implementing a new themeable based on the ggplot2 documentation,
    it is important to keep this in mind and reverse the order of the
    "inherits from" in the documentation.

    For example, to implement,

    - ``axis_title_x`` - ``x`` axis label (element_text;
      inherits from ``axis_title``)
    - ``axis_title_y`` - ``y`` axis label (element_text;
      inherits from ``axis_title``)


    You would have this implementation:

    ::

        class axis_title_x(themeable):
            ...

        class axis_title_y(themeable):
            ...

        class axis_title(axis_title_x, axis_title_y):
            ...


    If the superclasses fully implement the subclass, the body of the
    subclass should be "pass". Python(__mro__) will do the right thing.

    When a method does require implementation, call :python:`super()`
    then add the themeable's implementation to the axes.

    Notes
    -----
    A user should never create instances of class :class:`themeable` or
    subclasses of it.
    """

    order = 0
    properties: dict[str, Any]

    def __init__(self, theme_element: Any = None):
        self.theme_element = theme_element
        if isinstance(theme_element, element_base):
            self.properties = theme_element.properties
        else:
            # The specific themeable takes this value and
            # does stuff with rcParams or sets something
            # on some object attached to the axes/figure
            self.properties = {"value": theme_element}

        if isinstance(theme_element, element_blank):
            # TODO: Check if pyright complains about reassigning
            # functions too!
            self.apply_ax = self.blank_ax
            self.apply_figure = self.blank_figure

    @staticmethod
    def from_class_name(name: str, theme_element: Any) -> themeable:
        """
        Create an themeable by name

        Parameters
        ----------
        name : str
            Class name
        theme_element : element object
            A of the type required by the theme
            For lines, text and rects it should be one of:
            :class:`element_line`,
            :class:`element_rect`,
            :class:`element_text` or
            :class:`element_blank`

        Returns
        -------
        out : Themeable
        """
        msg = f"There no themeable element called: {name}"
        try:
            klass: Type[themeable] = themeable._registry[name]
        except KeyError:
            raise PlotnineError(msg)

        if not issubclass(klass, themeable):
            raise PlotnineError(msg)

        return klass(theme_element)

    @classmethod
    def registry(cls) -> Mapping[str, Any]:
        return themeable._registry

    def is_blank(self) -> bool:
        """
        Return True if theme_element is made of element_blank
        """
        return isinstance(self.theme_element, element_blank)

    def merge(self, other: themeable):
        """
        Merge properties of other into self

        Raises
        ------
        ValueError
            If any of the properties are blank
        """
        if self.is_blank() or other.is_blank():
            raise ValueError("Cannot merge if there is a blank.")
        else:
            self.properties.update(other.properties)

    def __eq__(self, other: Any) -> bool:
        "Mostly for unittesting."
        c1 = type(self) is type(other)
        c2: bool = self.properties == other.properties
        return c1 and c2

    @property
    def rcParams(self) -> dict[str, Any]:
        """
        Return themeables rcparams to an rcparam dict before plotting.

        Returns
        -------
        dict
            Dictionary of legal matplotlib parameters.

        This method should always call super(...).rcParams and
        update the dictionary that it returns with its own value, and
        return that dictionary.

        This method is called before plotting. It tends to be more
        useful for general themeables. Very specific themeables
        often cannot be be themed until they are created as a
        result of the plotting process.
        """
        return {}

    def apply(self, theme: Theme):
        """
        Called by the theme to apply the themeable

        Subclasses shouldn't have to override this method to customize.
        """
        self.apply_figure(theme.figure, theme._targets)
        for ax in theme.axs:
            self.apply_ax(ax)

    def apply_ax(self, ax: Axes):
        """
        Called after a chart has been plotted.

        Subclasses can override this method to customize the plot
        according to the theme.

        Parameters
        ----------
        ax : matplotlib.axes.Axes

        This method should be implemented as super(...).apply_ax()
        followed by extracting the portion of the axes specific to this
        themeable then applying the properties.
        """
        pass

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        """
        Apply theme to the figure

        Compared to :meth:`setup_figure`, this method is called
        after plotting and all the elements are drawn onto the
        figure.
        """
        pass

    def setup_figure(self, figure: Figure):
        """
        Apply theme to the figure

        Compared to :meth:`apply_figure`, this method is called
        before any plotting is done. This is necessary in some
        cases where the drawing functions need(or can make use of)
        this information.
        """
        pass

    def blank_ax(self, ax: Axes):
        """
        Blank out theme elements
        """
        pass

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        """
        Blank out elements on the figure
        """
        pass


class Themeables(Dict[str, themeable]):
    """
    Collection of themeables

    The key is the name of the class.
    """

    def update(self, other: Themeables):
        """
        Update themeables with those from `other`

        This method takes care of inserting the `themeable`
        into the underlying dictionary. Before doing the
        insertion, any existing themeables that will be
        affected by a new from `other` will either be merged
        or removed. This makes sure that a general themeable
        of type :class:`text` can be added to override an
        existing specific one of type :class:`axis_text_x`.
        """
        for new in other.values():
            new_key = new.__class__.__name__

            # 1st in the mro is self, the
            # last 2 are (themeable, object)
            for child in new.__class__.mro()[1:-2]:
                child_key = child.__name__
                try:
                    self[child_key].merge(new)
                except KeyError:
                    pass
                except ValueError:
                    # Blank child is will be overridden
                    del self[child_key]
            try:
                self[new_key].merge(new)
            except (KeyError, ValueError):
                # Themeable type is new or
                # could not merge blank element.
                self[new_key] = new

    def values(self) -> list[themeable]:
        """
        Return a list of themeables sorted in reverse based
        on the their depth in the inheritance hierarchy.

        The sorting is key when applying and merging the themeables
        so that they do not conflict i.e :class:`axis_line`
        applied before :class:`axis_line_x`.
        """

        def key(th: themeable) -> int:
            return len(th.__class__.__mro__)

        return sorted(dict.values(self), key=key, reverse=True)

    def property(self, name: str, key: str = "value") -> Any:
        """
        Get the value a specific themeable(s) property

        Themeables store theming attribute values in the
        :attr:`Themeable.properties` :class:`dict`. The goal
        of this method is to look a value from that dictionary,
        and fallback along the inheritance heirarchy of themeables.

        Parameters
        ----------
        name : str
            Themeable name
        key : str
            Property name to lookup

        Returns
        -------
        out : object
            Value

        Raises
        ------
        KeyError
            If key is in not in any of themeables
        """
        hlist = themeable._hierarchy[name]
        scalar = key == "value"
        for th in hlist:
            with suppress(KeyError):
                value = self[th].properties[key]
                if not scalar or value is not None:
                    return value

        msg = "'{}' is not in the properties of {} "
        raise KeyError(msg.format(key, hlist))

    def is_blank(self, name: str) -> bool:
        """
        Return True if the themeable *name* is blank

        If the *name* is not in the list of themeables then
        the lookup falls back to inheritance hierarchy.
        If none of the themeables are in the hierarchy are
        present, ``False`` is returned.

        Parameters
        ----------
        names : str
            Themeable, in order of most specific to most
            general.
        """
        for th in themeable._hierarchy[name]:
            with suppress(KeyError):
                return self[th].is_blank()

        return False


def _blankout_rect(rect: Patch):
    """
    Make rect invisible
    """
    # set_visible(False) does not clear the attributes
    rect.set_edgecolor("None")
    rect.set_facecolor("None")
    rect.set_linewidth(0)


# element_text themeables


class axis_title_x(themeable):
    """
    x axis label

    Parameters
    ----------
    theme_element : element_text
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        with suppress(KeyError):
            text = targets["axis_title_x"]
            text.set(**properties)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        with suppress(KeyError):
            text = targets["axis_title_x"]
            text.set_visible(False)


class axis_title_y(themeable):
    """
    y axis label

    Parameters
    ----------
    theme_element : element_text
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        with suppress(KeyError):
            text = targets["axis_title_y"]
            text.set(**properties)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            text = targets["axis_title_y"]
            text.set_visible(False)


class axis_title(axis_title_x, axis_title_y):
    """
    Axis labels

    Parameters
    ----------
    theme_element : element_text
    """

    pass


class legend_title(themeable):
    """
    Legend title

    Parameters
    ----------
    theme_element : element_text
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        with suppress(KeyError):
            textareas = targets["legend_title"]
            for ta in textareas:
                ta._text.set(**properties)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            textareas = targets["legend_title"]
            for ta in textareas:
                ta.set_visible(False)


class legend_text_legend(themeable):
    """
    Legend text for the common legend

    Parameters
    ----------
    theme_element : element_text

    Notes
    -----
    This themeable exists mainly to cater for differences
    in how the text is aligned compared to the colorbar.
    Unless you experience those alignment issues (i.e when
    using parameters **va** or **ha**), you should use
    :class:`legend_text`.
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        with suppress(KeyError):
            texts = targets["legend_text_legend"]
            for text in texts:
                if not hasattr(text, "_x"):  # textarea
                    text = text._text
                text.set(**properties)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            texts = targets["legend_text_legend"]
            for text in texts:
                text.set_visible(False)


class legend_text_colorbar(themeable):
    """
    Colorbar text

    Parameters
    ----------
    theme_element : element_text

    Notes
    -----
    This themeable exists mainly to cater for differences
    in how the text is aligned compared to the entry based
    legend. Unless you experience those alignment issues
    (i.e when using parameters **va** or **ha**), you should
    use :class:`legend_text`.
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        with suppress(KeyError):
            texts = targets["legend_text_colorbar"]
            for text in texts:
                if not hasattr(text, "_x"):  # textarea
                    text = text._text
                text.set(**properties)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            texts = targets["legend_text_colorbar"]
            for text in texts:
                text.set_visible(False)


legend_text_colourbar = legend_text_colorbar


class legend_text(legend_text_legend, legend_text_colorbar):
    """
    Legend text

    Parameters
    ----------
    theme_element : element_text
    """


class plot_title(themeable):
    """
    Plot title

    Parameters
    ----------
    theme_element : element_text
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        with suppress(KeyError):
            text = targets["plot_title"]
            text.set(**properties)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            text = targets["plot_title"]
            text.set_visible(False)


class plot_subtitle(themeable):
    """
    Plot subtitle

    Parameters
    ----------
    theme_element : element_text
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        with suppress(KeyError):
            text = targets["plot_subtitle"]
            text.set(**properties)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            text = targets["plot_subtitle"]
            text.set_visible(False)


class plot_caption(themeable):
    """
    Plot caption

    Parameters
    ----------
    theme_element : element_text
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        with suppress(KeyError):
            text = targets["plot_caption"]
            text.set(**properties)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            text = targets["plot_caption"]
            text.set_visible(False)


class strip_text_x(themeable):
    """
    Facet labels along the horizontal axis

    Parameters
    ----------
    theme_element : element_text
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        with suppress(KeyError):
            texts = targets["strip_text_x"]
            for text in texts:
                text.set(**properties)

        with suppress(KeyError):
            rects = targets["strip_background_x"]
            for rect in rects:
                rect.set_visible(True)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            texts = targets["strip_text_x"]
            for text in texts:
                text.set_visible(False)

        with suppress(KeyError):
            rects = targets["strip_background_x"]
            for rect in rects:
                rect.set_visible(False)


class strip_text_y(themeable):
    """
    Facet labels along the vertical axis

    Parameters
    ----------
    theme_element : element_text
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        with suppress(KeyError):
            texts = targets["strip_text_y"]
            for text in texts:
                text.set(**properties)

        with suppress(KeyError):
            rects = targets["strip_background_y"]
            for rect in rects:
                rect.set_visible(True)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            texts = targets["strip_text_y"]
            for text in texts:
                text.set_visible(False)

        with suppress(KeyError):
            rects = targets["strip_background_y"]
            for rect in rects:
                rect.set_visible(False)


class strip_text(strip_text_x, strip_text_y):
    """
    Facet labels along both axes

    Parameters
    ----------
    theme_element : element_text
    """

    pass


class title(axis_title, legend_title, plot_title, plot_subtitle, plot_caption):
    """
    All titles on the plot

    Parameters
    ----------
    theme_element : element_text
    """

    pass


class axis_text_x(themeable):
    """
    x-axis tick labels

    Parameters
    ----------
    theme_element : element_text
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        labels = ax.get_xticklabels()  # pyright: ignore
        for l in labels:
            l.set(**properties)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.xaxis.set_tick_params(
            which="both", labelbottom=False, labeltop=False
        )


class axis_text_y(themeable):
    """
    y-axis tick labels

    Parameters
    ----------
    theme_element : element_text
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        properties = self.properties.copy()
        with suppress(KeyError):
            del properties["margin"]
        labels = ax.get_yticklabels()  # pyright: ignore
        for l in labels:
            l.set(**properties)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.yaxis.set_tick_params(
            which="both", labelleft=False, labelright=False
        )


class axis_text(axis_text_x, axis_text_y):
    """
    Axis tick labels

    Parameters
    ----------
    theme_element : element_text
    """

    pass


class text(axis_text, legend_text, strip_text, title):
    """
    All text elements in the plot

    Parameters
    ----------
    theme_element : element_text
    """

    @property
    def rcParams(self) -> dict[str, Any]:
        rcParams = super().rcParams

        family = self.properties.get("family")
        style = self.properties.get("style")
        weight = self.properties.get("weight")
        size = self.properties.get("size")
        color = self.properties.get("color")

        if family:
            rcParams["font.family"] = family
        if style:
            rcParams["font.style"] = style
        if weight:
            rcParams["font.weight"] = weight
        if size:
            rcParams["font.size"] = size
            rcParams["xtick.labelsize"] = size
            rcParams["ytick.labelsize"] = size
            rcParams["legend.fontsize"] = size
        if color:
            rcParams["text.color"] = color

        return rcParams


# element_line themeables


class axis_line_x(themeable):
    """
    x-axis line

    Parameters
    ----------
    theme_element : element_line
    """

    position = "bottom"

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        with suppress(KeyError):
            del self.properties["solid_capstyle"]

        ax.spines["top"].set_visible(False)
        ax.spines["bottom"].set(**self.properties)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.spines["top"].set_visible(False)
        ax.spines["bottom"].set_visible(False)


class axis_line_y(themeable):
    """
    y-axis line

    Parameters
    ----------
    theme_element : element_line
    """

    position = "left"

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        with suppress(KeyError):
            del self.properties["solid_capstyle"]

        ax.spines["right"].set_visible(False)
        ax.spines["left"].set(**self.properties)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.spines["left"].set_visible(False)
        ax.spines["right"].set_visible(False)


class axis_line(axis_line_x, axis_line_y):
    """
    x & y axis lines

    Parameters
    ----------
    theme_element : element_line
    """

    pass


class axis_ticks_minor_x(themeable):
    """
    x-axis tick lines

    Parameters
    ----------
    theme_element : element_line
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)

        d = deepcopy(self.properties)
        with suppress(KeyError):
            d["markeredgewidth"] = d.pop("linewidth")
        with suppress(KeyError):
            d["markeredgecolor"] = d.pop("color")

        for tick in ax.xaxis.get_minor_ticks():
            tick.tick1line.set(**d)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.xaxis.set_tick_params(which="minor", bottom=False)


class axis_ticks_minor_y(themeable):
    """
    y-axis minor tick lines

    Parameters
    ----------
    theme_element : element_line
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)

        d = deepcopy(self.properties)
        with suppress(KeyError):
            d["markeredgewidth"] = d.pop("linewidth")
        with suppress(KeyError):
            d["markeredgecolor"] = d.pop("color")

        for tick in ax.yaxis.get_minor_ticks():
            tick.tick1line.set(**d)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.yaxis.set_tick_params(which="minor", left=False)


class axis_ticks_major_x(themeable):
    """
    x-axis major tick lines

    Parameters
    ----------
    theme_element : element_line
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)

        d = deepcopy(self.properties)
        del d["visible"]
        with suppress(KeyError):
            d["markeredgewidth"] = d.pop("linewidth")
        with suppress(KeyError):
            d["markeredgecolor"] = d.pop("color")

        for tick in ax.xaxis.get_major_ticks():
            tick.tick1line.set(**d)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.xaxis.set_tick_params(which="major", bottom=False)


class axis_ticks_major_y(themeable):
    """
    y-axis major tick lines

    Parameters
    ----------
    theme_element : element_line
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)

        d = deepcopy(self.properties)
        del d["visible"]
        with suppress(KeyError):
            d["markeredgewidth"] = d.pop("linewidth")
        with suppress(KeyError):
            d["markeredgecolor"] = d.pop("color")

        for tick in ax.yaxis.get_major_ticks():
            tick.tick1line.set(**d)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.yaxis.set_tick_params(which="major", left=False)


class axis_ticks_major(axis_ticks_major_x, axis_ticks_major_y):
    """
    x & y axis major tick lines

    Parameters
    ----------
    theme_element : element_line
    """

    pass


class axis_ticks_minor(axis_ticks_minor_x, axis_ticks_minor_y):
    """
    x & y axis minor tick lines

    Parameters
    ----------
    theme_element : element_line
    """

    pass


class axis_ticks(axis_ticks_major, axis_ticks_minor):
    """
    x & y major and minor axis tick lines

    Parameters
    ----------
    theme_element : element_line
    """

    pass


class panel_grid_major_x(themeable):
    """
    Vertical major grid lines

    Parameters
    ----------
    theme_element : element_line
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        ax.xaxis.grid(which="major", **self.properties)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.grid(False, which="major", axis="x")


class panel_grid_major_y(themeable):
    """
    Horizontal major grid lines

    Parameters
    ----------
    theme_element : element_line
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        ax.yaxis.grid(which="major", **self.properties)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.grid(False, which="major", axis="y")


class panel_grid_minor_x(themeable):
    """
    Vertical minor grid lines

    Parameters
    ----------
    theme_element : element_line
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        ax.xaxis.grid(which="minor", **self.properties)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.grid(False, which="minor", axis="x")


class panel_grid_minor_y(themeable):
    """
    Horizontal minor grid lines

    Parameters
    ----------
    theme_element : element_line
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        ax.yaxis.grid(which="minor", **self.properties)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.grid(False, which="minor", axis="y")


class panel_grid_major(panel_grid_major_x, panel_grid_major_y):
    """
    Major grid lines

    Parameters
    ----------
    theme_element : element_line
    """

    pass


class panel_grid_minor(panel_grid_minor_x, panel_grid_minor_y):
    """
    Minor grid lines

    Parameters
    ----------
    theme_element : element_line
    """

    pass


class panel_grid(panel_grid_major, panel_grid_minor):
    """
    Grid lines

    Parameters
    ----------
    theme_element : element_line
    """

    pass


class line(axis_line, axis_ticks, panel_grid):
    """
    All line elements

    Parameters
    ----------
    theme_element : element_line
    """

    @property
    def rcParams(self) -> dict[str, Any]:
        rcParams = super().rcParams
        color = self.properties.get("color")
        linewidth = self.properties.get("linewidth")
        linestyle = self.properties.get("linestyle")
        d = {}

        if color:
            d["axes.edgecolor"] = color
            d["xtick.color"] = color
            d["ytick.color"] = color
            d["grid.color"] = color
        if linewidth:
            d["axes.linewidth"] = linewidth
            d["xtick.major.width"] = linewidth
            d["xtick.minor.width"] = linewidth
            d["ytick.major.width"] = linewidth
            d["ytick.minor.width"] = linewidth
            d["grid.linewidth"] = linewidth
        if linestyle:
            d["grid.linestyle"] = linestyle

        rcParams.update(d)
        return rcParams


# element_rect themeables


class legend_key(themeable):
    """
    Legend key background

    Parameters
    ----------
    theme_element : element_rect
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        with suppress(KeyError):
            # list of lists
            all_drawings = targets["legend_key"]
            for drawings in all_drawings:
                for da in drawings:
                    da.patch.set(**self.properties)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            # list of lists
            all_drawings = targets["legend_key"]
            for drawings in all_drawings:
                for da in drawings:
                    _blankout_rect(da.patch)


class legend_background(themeable):
    """
    Legend background

    Parameters
    ----------
    theme_element : element_rect
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        # anchored offset box
        with suppress(KeyError):
            aob = targets["legend_background"]
            aob.patch.set(**self.properties)
            if self.properties:
                aob._drawFrame = True
                # some small sensible padding
                if not aob.pad:
                    aob.pad = 0.2

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            aob = targets["legend_background"]
            _blankout_rect(aob.patch)


class legend_box_background(themeable):
    """
    Legend box background

    Parameters
    ----------
    theme_element : element_rect

    Notes
    -----
    Not Implemented. We would have to place the outermost
    VPacker/HPacker boxes that hold the individual legends
    onto an object that has a patch.
    """


class panel_background(themeable):
    """
    Panel background

    Parameters
    ----------
    theme_element : element_rect
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        d = deepcopy(self.properties)
        if "facecolor" in d and "alpha" in d:
            d["facecolor"] = to_rgba(d["facecolor"], d["alpha"])
            del d["alpha"]
        ax.patch.set(**d)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        _blankout_rect(ax.patch)


class panel_border(themeable):
    """
    Panel border

    Parameters
    ----------
    theme_element : element_rect
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        d = deepcopy(self.properties)
        # Be lenient, if using element_line
        with suppress(KeyError):
            d["edgecolor"] = d.pop("color")

        with suppress(KeyError):
            del d["facecolor"]

        if "edgecolor" in d and "alpha" in d:
            d["edgecolor"] = to_rgba(d["edgecolor"], d["alpha"])
            del d["alpha"]

        ax.patch.set(**d)

    def blank_ax(self, ax: Axes):
        super().blank_ax(ax)
        ax.patch.set_linewidth(0)


class plot_background(themeable):
    """
    Plot background

    Parameters
    ----------
    theme_element : element_rect
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        figure.patch.set(**self.properties)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        _blankout_rect(figure.patch)


class strip_background_x(themeable):
    """
    Horizontal facet label background

    Parameters
    ----------
    theme_element : element_rect
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        with suppress(KeyError):
            bboxes = targets["strip_background_x"]
            for bbox in bboxes:
                bbox.set(**self.properties)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            rects = targets["strip_background_x"]
            for rect in rects:
                _blankout_rect(rect)


class strip_background_y(themeable):
    """
    Vertical facet label background

    Parameters
    ----------
    theme_element : element_rect
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        super().apply_figure(figure, targets)
        with suppress(KeyError):
            bboxes = targets["strip_background_y"]
            for bbox in bboxes:
                bbox.set(**self.properties)

    def blank_figure(self, figure: Figure, targets: dict[str, Any]):
        super().blank_figure(figure, targets)
        with suppress(KeyError):
            rects = targets["strip_background_y"]
            for rect in rects:
                _blankout_rect(rect)


class strip_background(strip_background_x, strip_background_y):
    """
    Facet label background

    Parameters
    ----------
    theme_element : element_rect
    """

    pass


class rect(
    legend_key,
    legend_background,
    panel_background,
    panel_border,
    plot_background,
    strip_background,
):
    """
    All rectangle elements

    Parameters
    ----------
    theme_element : element_rect
    """

    pass


# value base themeables


class axis_ticks_length_major(themeable):
    """
    Axis major-tick length

    Parameters
    ----------
    theme_element : float
        Value in points.
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        ax.xaxis.set_tick_params(which="major", size=self.properties["value"])
        ax.yaxis.set_tick_params(which="major", size=self.properties["value"])


class axis_ticks_length_minor(themeable):
    """
    Axis minor-tick length

    Parameters
    ----------
    theme_element : float
        Value in points.
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        ax.xaxis.set_tick_params(which="minor", size=self.properties["value"])
        ax.yaxis.set_tick_params(which="minor", size=self.properties["value"])


class axis_ticks_length(axis_ticks_length_major, axis_ticks_length_minor):
    """
    Axis tick length

    Parameters
    ----------
    theme_element : float
        Value in points.
    """

    pass


class axis_ticks_pad_major(themeable):
    """
    Axis major-tick padding

    Parameters
    ----------
    theme_element : float
        Value in points.
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        val = self.properties["value"]

        for t in ax.xaxis.get_major_ticks():
            t.set_pad(val)

        for t in ax.yaxis.get_major_ticks():
            t.set_pad(val)


class axis_ticks_pad_minor(themeable):
    """
    Axis minor-tick padding

    Parameters
    ----------
    theme_element : float
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        val = self.properties["value"]

        for t in ax.xaxis.get_minor_ticks():
            t.set_pad(val)

        for t in ax.yaxis.get_minor_ticks():
            t.set_pad(val)


class axis_ticks_pad(axis_ticks_pad_major, axis_ticks_pad_minor):
    """
    Axis tick padding

    Parameters
    ----------
    theme_element : float
        Value in points.
    """

    pass


class axis_ticks_direction_x(themeable):
    """
    x-axis tick direction

    Parameters
    ----------
    theme_element : str in ``['in', 'out', 'inout']``
        - ``in`` - ticks inside the panel
        - ``out`` - ticks outside the panel
        - ``inout`` - ticks inside and outside the panel
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        ax.xaxis.set_tick_params(
            which="major", tickdir=self.properties["value"]
        )


class axis_ticks_direction_y(themeable):
    """
    y-axis tick direction

    Parameters
    ----------
    theme_element : str in ``['in', 'out', 'inout']``
        - ``in`` - ticks inside the panel
        - ``out`` - ticks outside the panel
        - ``inout`` - ticks inside and outside the panel
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        ax.yaxis.set_tick_params(
            which="major", tickdir=self.properties["value"]
        )


class axis_ticks_direction(axis_ticks_direction_x, axis_ticks_direction_y):
    """
    axis tick direction

    Parameters
    ----------
    theme_element : {'in', 'out', 'inout'}
        - ``in`` - ticks inside the panel
        - ``out`` - ticks outside the panel
        - ``inout`` - ticks inside and outside the panel
    """

    pass


class panel_spacing_x(themeable):
    """
    Horizontal spacing between the facet panels

    Parameters
    ----------
    theme_element : float
        Size as a fraction of the figure width.
    """

    pass


class panel_spacing_y(themeable):
    """
    Vertical spacing between the facet panels

    Parameters
    ----------
    theme_element : float
        Size as a fraction of the figure width.

    Notes
    -----
    It is deliberate to have the vertical spacing be a fraction of
    the width. That means that when :class:`panel_spacing_x` is the
    equal :class:`panel_spacing_x`, the spaces in both directions
    will be equal.
    """

    pass


class panel_spacing(panel_spacing_x, panel_spacing_y):
    """
    Spacing between the facet panels

    Parameters
    ----------
    theme_element : float
        Size in inches of the space between the facet panels
    """

    pass


# TODO: Distinct margins in all four directions
class plot_margin(themeable):
    """
    Plot Margin

    Parameters
    ----------
    theme_element : float
        Must be in the [0, 1] range. It is specified
        as a fraction of the figure width and figure
        height.
    """


class panel_ontop(themeable):
    """
    Place panel background & gridlines over/under the data layers

    Parameters
    ----------
    theme_element : bool
        Default is False.
    """

    def apply_ax(self, ax: Axes):
        super().apply_ax(ax)
        ax.set_axisbelow(not self.properties["value"])


class aspect_ratio(themeable):
    """
    Aspect ratio of the panel(s)

    Parameters
    ----------
    theme_element : float
        `panel_height / panel_width`

    Notes
    -----
    For a fixed relationship between the ``x`` and ``y`` scales,
    use :class:`~plotnine.coords.coord_fixed`.
    """


class dpi(themeable):
    """
    DPI with which to draw/save the figure

    Parameters
    ----------
    theme_element : int
    """

    # fig.set_dpi does not work
    # https://github.com/matplotlib/matplotlib/issues/24644

    @property
    def rcParams(self) -> dict[str, Any]:
        rcParams = super().rcParams
        rcParams["figure.dpi"] = self.properties["value"]
        return rcParams


class figure_size(themeable):
    """
    Figure size in inches

    Parameters
    ----------
    theme_element : tuple
        (width, height) in inches
    """

    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        figure.set_size_inches(self.properties["value"])


class legend_box(themeable):
    """
    How to box up multiple legends

    Parameters
    ----------
    theme_element : str in ``['vertical', 'horizontal']``
        Whether to stack up the legends vertically or
        horizontally.
    """


class legend_box_margin(themeable):
    """
    Padding between the legends and the box

    Parameters
    ----------
    theme_element : int
        Value in points.
    """


class legend_box_just(themeable):
    """
    Justification of legend boxes

    Parameters
    ----------
    theme_element : str
        One of *left*, *right*, *center*, *top* or *bottom*
        depending the value of :class:`legend_box`.
    """


class legend_direction(themeable):
    """
    Layout items in the legend

    Parameters
    ----------
    theme_element : str in ``['vertical', 'horizontal']``
        Vertically or horizontally
    """


class legend_key_width(themeable):
    """
    Legend key background width

    Parameters
    ----------
    theme_element : float
        Value in points
    """


class legend_key_height(themeable):
    """
    Legend key background height

    Parameters
    ----------
    theme_element : float
        Value in points.
    """


class legend_key_size(legend_key_width, legend_key_height):
    """
    Legend key background width and height

    Parameters
    ----------
    theme_element : float
        Value in points.
    """


class legend_margin(themeable):
    """
    Padding between the legend and the inner box

    Parameters
    ----------
    theme_element : float
        Value in points
    """


class legend_box_spacing(themeable):
    """
    Spacing between the legend and the plotting area

    Parameters
    ----------
    theme_element : float
        Value in points.
    """


class legend_spacing(themeable):
    """
    Spacing between two adjacent legends

    Parameters
    ----------
    theme_element : float
        Value in points.
    """


class legend_position(themeable):
    """
    Location of legend

    Parameters
    ----------
    theme_element : str or tuple
        If a string it should be one of *right*, *left*, *top*
        *bottom* or *none*. If a tuple, it should be two floats each
        in the approximate range [0, 1]. The tuple specifies the
        location of the legend in screen coordinates.
    """


class legend_title_align(themeable):
    """
    Alignment of legend title

    Parameters
    ----------
    theme_element : str or tuple
        If a string it should be one of *right*, *left*, *center*,
        *top* or *bottom*.
    """


class legend_entry_spacing_x(themeable):
    """
    Horizontal spacing between two entries in a legend

    Parameters
    ----------
    theme_element : float
        Size in points
    """


class legend_entry_spacing_y(themeable):
    """
    Vertical spacing between two entries in a legend

    Parameters
    ----------
    theme_element : float
        Size in points
    """


class legend_entry_spacing(legend_entry_spacing_x, legend_entry_spacing_y):
    """
    Spacing between two entries in a legend

    Parameters
    ----------
    theme_element : float
        Size in points
    """


class strip_align_x(themeable):
    """
    Vertical alignment of the strip & its background w.r.t the panel border

    Parameters
    ----------
    theme_element : float
        Value as a proportion of the strip size. A good value
        should be the range :math:`[-1, 0.5]`. A negative value
        puts the strip inside the axes. A positive value creates
        a margin between the strip and the axes. `0` puts the
        strip on top of the panels.
    """


class strip_align_y(themeable):
    """
    Horizontal alignment of the strip & its background w.r.t the panel border

    Parameters
    ----------
    theme_element : float
        Value as a proportion of the strip size. A good value
        should be the range :math:`[-1, 0.5]`. A negative value
        puts the strip inside the axes. A positive value creates
        a margin between the strip and the axes. `0` puts the
        strip exactly beside the panels.
    """


class strip_align(strip_align_x, strip_align_y):
    """
    Alignment of the strip & its background w.r.t the panel border

    Parameters
    ----------
    theme_element : float
        Value as a proportion of the strip text size. A good value
        should be the range :math:`[-1, 0.5]`. A negative value
        puts the strip inside the axes and a positive value
        creates a space between the strip and the axes.
    """


# Deprecated


class subplots_adjust(themeable):
    def apply_figure(self, figure: Figure, targets: dict[str, Any]):
        warn(
            "You no longer need to use subplots_adjust to make space for "
            "the legend or text around the panels. This paramater will be "
            "removed in a future version. You can still use 'plot_margin' "
            "'panel_spacing' for your other spacing needs.",
            FutureWarning,
        )



from __future__ import annotations

import typing

import numpy as np
import pandas as pd

from ..doctools import document
from ..utils import SIZE_FACTOR, interleave, make_line_segments, to_rgba
from .geom import geom
from .geom_path import geom_path

if typing.TYPE_CHECKING:
    from typing import Any

    from plotnine.iapi import panel_view
    from plotnine.typing import (
        Axes,
        Coord,
    )


@document
class geom_segment(geom):
    """
    Line segments

    {usage}

    Parameters
    ----------
    {common_parameters}
    lineend : str (default: butt)
        Line end style, of of *butt*, *round* or *projecting.*
        This option is applied for solid linetypes.
    arrow : plotnine.geoms.geom_path.arrow (default: None)
        Arrow specification. Default is no arrow.

    See Also
    --------
    plotnine.geoms.geom_path.arrow : for adding arrowhead(s)
        to segments.
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "black",
        "linetype": "solid",
        "size": 0.5,
    }
    REQUIRED_AES = {"x", "y", "xend", "yend"}
    NON_MISSING_AES = {"linetype", "size", "shape"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "lineend": "butt",
        "arrow": None,
    }

    draw_legend = staticmethod(geom_path.draw_legend)

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        from matplotlib.collections import LineCollection

        data = coord.transform(data, panel_params)
        data["size"] *= SIZE_FACTOR
        color = to_rgba(data["color"], data["alpha"])

        # start point -> end point, sequence of xy points
        # from which line segments are created
        x = interleave(data["x"], data["xend"])
        y = interleave(data["y"], data["yend"])
        segments = make_line_segments(x, y, ispath=False)
        coll = LineCollection(
            segments,
            edgecolor=color,
            linewidth=data["size"],
            linestyle=data["linetype"][0],
            zorder=params["zorder"],
            rasterized=params["raster"],
        )
        ax.add_collection(coll)

        if "arrow" in params and params["arrow"]:
            adata = pd.DataFrame(index=range(len(data) * 2))
            idx = np.arange(1, len(data) + 1)
            adata["group"] = np.hstack([idx, idx])
            adata["x"] = np.hstack([data["x"], data["xend"]])
            adata["y"] = np.hstack([data["y"], data["yend"]])
            other = ["color", "alpha", "size", "linetype"]
            for param in other:
                adata[param] = np.hstack([data[param], data[param]])

            params["arrow"].draw(
                adata, panel_params, coord, ax, constant=False, **params
            )



from __future__ import annotations

import typing

from ..exceptions import PlotnineError
from ..utils import Registry

if typing.TYPE_CHECKING:
    from typing import Any, Literal, Optional

    import pandas as pd

    from plotnine.typing import Scale, Theme


class guide(metaclass=Registry):
    """
    Base class for all guides

    Parameters
    ----------
    title : str | None
        Title of the guide. If ``None``, title is not shown.
        Default is the name of the aesthetic or the name
        specified using :class:`~plotnine.components.labels.lab`
    title_position : str in ``['top', 'bottom', 'left', 'right']``
        Position of title
    title_theme : element_text
        Control over the title theme.
        Default is to use ``legend_title`` in a theme.
    title_hjust : float
        Horizontal justification of title text.
    title_vjust : float
        Vertical justification of title text.
    title_separation : float
        Separation between the title text and the colorbar.
        Value is in pixels.
    label : bool
        Whether to show labels
    label_position : str in ``['top', 'bottom', 'left', 'right']``
        Position of the labels.
        The defaults are ``'bottom'`` for a horizontal guide and
        '``right``' for a vertical guide.
    label_theme : element_text
        Control over the label theme.
        Default is to use ``legend_text`` in a theme.
    label_hjust : float
        Horizontal justification of label text.
    label_vjust : float
        Vertical justification of label text.
    label_separation : float
        Separation between the label text and the colorbar.
        Value is in pixels.
    direction : str in ``['horizontal', 'vertical']``
        Direction of the guide.
    default_unit : str
        Unit for ``keywidth`` and ``keyheight``
    override_aes : dict
        Aesthetic parameters of legend key.
    reverse : bool
        Whether to reverse the order of the legends.
    order : int
        Order of this guide among multiple guides.
        Should be in the range [0, 99]. Default is ``0``.

    Notes
    -----
    At the moment not all parameters have been fully implemented.
    """

    __base__ = True

    # Must be updated before the draw method is called
    theme: Theme
    key: pd.DataFrame

    # title
    title: str
    title_position: Literal["top", "bottom", "left", "right"]
    title_theme = None
    title_hjust: Optional[float] = None
    title_vjust: Optional[float] = None

    # label
    label = True
    label_position: Literal["top", "bottom", "left", "right"]
    label_theme = None
    label_hjust: Optional[float] = None
    label_vjust: Optional[float] = None

    # general
    direction: Optional[Literal["horizontal", "vertical"]] = None
    default_unit = "line"
    override_aes: dict[str, Any] = {}
    reverse = False
    order = 0

    # Aesthetics covered by the guide
    available_aes: set[str] = set()

    def __init__(self, **kwargs):
        no_default = {"title", "title_position", "label_position"}
        for k, v in kwargs.items():
            if hasattr(self, k) or k in no_default:
                setattr(self, k, v)
            else:
                tpl = "{} does not undestand attribute '{}'"
                raise PlotnineError(tpl.format(self.__class__.__name__, k))

    def _set_defaults(self, theme: Theme):
        """
        Set configuration parameters for drawing guide
        """
        self.theme = theme
        valid_locations = {"top", "bottom", "left", "right"}
        _property = theme.themeables.property
        margin_location_lookup = {
            # Where to put the margin between the legend and
            # the axes. Depends on the location of the legend
            "t": "b",
            "b": "t",
            "l": "r",
            "r": "l",
        }

        # label position
        if not hasattr(self, "label_position"):
            self.label_position = "right"

        if self.label_position not in valid_locations:
            msg = "label position '{}' is invalid"
            raise PlotnineError(msg.format(self.label_position))

        # label margin
        # legend_text_legend or legend_text_colorbar
        _legend_type = self.__class__.__name__.split("_")[-1]
        name = "legend_text_{}".format(_legend_type)
        loc = margin_location_lookup[self.label_position[0]]
        margin = _property(name, "margin")
        self._label_margin = margin.get_as(loc, "pt")

        # direction of guide
        if self.direction is None:
            if self.label_position in ("right", "left"):
                self.direction = "vertical"
            else:
                self.direction = "horizontal"

        # title position
        if not hasattr(self, "title_position"):
            if self.direction == "vertical":
                self.title_position = "top"
            elif self.direction == "horizontal":
                self.title_position = "left"
        if self.title_position not in valid_locations:
            msg = "legend title position '{}' is invalid"
            raise PlotnineError(msg.format(self.title_position))

        # title alignment
        self._title_align = _property("legend_title_align")
        if self._title_align == "auto":
            if self.direction == "vertical":
                self._title_align = "left"
            else:
                self._title_align = "center"

        # by default, direction of each guide depends on
        # the position all the guides
        position = _property("legend_position")
        self.direction = _property("legend_direction")
        if self.direction == "auto":
            if position in ("right", "left"):  # default
                self.direction = "vertical"
            else:
                self.direction = "horizontal"

        # title margin
        loc = margin_location_lookup[self.title_position[0]]
        margin = _property("legend_title", "margin")
        self._title_margin = margin.get_as(loc, "pt")

        # legend_margin
        self._legend_margin = _property("legend_margin")

        # legend_entry_spacing
        self._legend_entry_spacing_x = _property("legend_entry_spacing_x")
        self._legend_entry_spacing_y = _property("legend_entry_spacing_y")

    def legend_aesthetics(self, layer, plot):
        """
        Return the aesthetics that contribute to the legend

        Parameters
        ----------
        layer : Layer
            Layer whose legend is to be drawn
        plot : ggplot
            Plot object

        Returns
        -------
        matched : list
            List of the names of the aethetics that contribute
            to the legend.
        """
        l = layer
        legend_ae = set(self.key.columns) - {"label"}
        all_ae = (
            l.mapping.keys()
            | (plot.mapping if l.inherit_aes else set())
            | l.stat.DEFAULT_AES.keys()
        )
        geom_ae = l.geom.REQUIRED_AES | l.geom.DEFAULT_AES.keys()
        matched = all_ae & geom_ae & legend_ae
        matched = list(matched - set(l.geom.aes_params))
        return matched

    def train(
        self, scale: Scale, aesthetic: Optional[str] = None
    ) -> guide | None:
        """
        Create the key for the guide

        Returns guide if training is successful
        """
        pass



import numpy as np
import pandas as pd

from plotnine import (
    aes,
    annotate,
    facet_grid,
    facet_wrap,
    geom_abline,
    geom_path,
    geom_point,
    ggplot,
)
from plotnine.data import mpg, mtcars

n = 10
df = pd.DataFrame(
    {
        "x": range(n),
        "y": range(n),
        "var1": np.repeat(range(n // 2), 2),
        "var2": np.tile(["a", "b"], n // 2),
    }
)
df["class"] = df["var1"]  # python keyword as column
df["g"] = df["var1"]  # variable as a column

g = ggplot(df, aes("x", "y")) + geom_point(
    aes(color="factor(var1)"), size=5, show_legend=False
)


# facet_wrap


def test_facet_wrap_one_var():
    p = g + facet_wrap("~var1")
    p2 = g + facet_wrap("~class")  # python keyword in formula
    p3 = g + facet_wrap("~g")  # variable in formula
    assert p == "facet_wrap_one_var"
    assert p2 == "facet_wrap_one_var"
    assert p3 == "facet_wrap_one_var"


def test_facet_wrap_expression():
    p = g + facet_wrap("pd.cut(var1, (0, 2, 4), include_lowest=True)")
    assert p == "facet_wrap_expression"


def test_facet_wrap_two_vars():
    p = g + facet_wrap("~var1+var2")
    p2 = g + facet_wrap("~class+var2")  # python keyword in formula
    assert p == "facet_wrap_two_vars"
    assert p2 == "facet_wrap_two_vars"


def test_facet_wrap_label_both():
    p = g + facet_wrap("~var1+var2", labeller="label_both")
    assert p == "facet_wrap_label_both"


def test_facet_wrap_not_as_table():
    p = g + facet_wrap("~var1", as_table=False)
    assert p == "facet_wrap_not_as_table"


def test_facet_wrap_direction_v():
    p = g + facet_wrap("~var1", dir="v")
    assert p == "facet_wrap_direction_v"


def test_facet_wrap_not_as_table_direction_v():
    p = g + facet_wrap("~var1", as_table=False, dir="v")
    assert p == "facet_wrap_not_as_table_direction_v"


# facet_grid


def test_facet_grid_one_by_one_var():
    p = g + facet_grid("var1~var2")
    p2 = g + facet_grid("class~var2")  # python keyword in formula
    assert p == "facet_grid_one_by_one_var"
    assert p2 == "facet_grid_one_by_one_var"


def test_facet_grid_expression():
    p = g + facet_grid(
        ["var2", "pd.cut(var1, (0, 2, 4), include_lowest=True)"]
    )
    assert p == "facet_grid_expression"


def test_facet_grid_margins():
    p = g + facet_grid("var1~var2", margins=True)
    assert p == "facet_grid_margins"


def test_facet_grid_scales_free_y():
    p = g + facet_grid("var1>2 ~ x%2", scales="free_y")
    assert p == "facet_grid_scales_free_y"


def test_facet_grid_formula_with_dot():
    p = g + facet_grid(". ~ var1>2")
    assert p == "facet_grid_formula_with_dot"


def test_facet_grid_formula_without_dot():
    p = g + facet_grid("~var1>2")
    assert p == "facet_grid_formula_with_dot"


def test_facet_grid_scales_free_x():
    p = g + facet_grid("var1>2 ~ x%2", scales="free_x")
    assert p == "facet_grid_scales_free_x"


def test_facet_grid_drop_false():
    df = mpg.copy()
    df["drv"] = pd.Categorical(df["drv"], ["4", "f", "r", "Q"])

    p = (
        ggplot(df, aes(x="displ", y="hwy"))
        + geom_point()
        + facet_grid("drv ~ .", drop=False)
    )
    assert p == "facet_grid_drop_false"


def test_facet_grid_space_ratios():
    p = (
        ggplot(mtcars, aes("wt", "mpg"))
        + geom_point()
        + facet_grid("am ~ vs", space={"y": [1, 2], "x": [1, 2]})
    )
    assert p == "facet_grid_space_ratios"


# Edge cases


def test_non_mapped_facetting():
    p = g + geom_abline(intercept=0, slope=1, size=1) + facet_wrap("var1")
    assert p == "non_mapped_facetting"


def test_dir_v_ncol():
    p = (
        ggplot(mpg)
        + aes(x="displ", y="hwy")
        + facet_wrap("class", dir="v", ncol=4, as_table=False)
        + geom_point()
    )
    assert p == "dir_v_ncol"


def test_variable_and_annotate():
    p = (
        g
        + annotate("point", x=4.5, y=5.5, color="cyan", size=10)
        + facet_wrap("~g")
    )
    assert p == "variable_and_annotate"


def test_manual_mapping_with_lists():
    p = (
        g
        + annotate(
            "abline", intercept=[4, 10], slope=[-1, -1], color=["red", "blue"]
        )
        + facet_wrap("var1")
    )
    assert p == "manual_mapping_with_lists"


def test_array_mapping_and_evaluation():
    # GH548
    # When we map to array/series of values, the aes evaluation
    # should not mix up the values.
    df = pd.DataFrame({"x": range(12), "y": range(12), "g": list("abcd") * 3})

    # should be the same as if color='g'
    p = (
        ggplot(df, aes("x", "y", color=df["g"]))
        + geom_point(size=4)
        + geom_path()
        + facet_wrap("~g")
    )
    assert p == "array_mapping_and_evaluation"



import warnings

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineWarning
from .smoothers import predictdf
from .stat import stat


@document
class stat_smooth(stat):
    """
    Calculate a smoothed conditional mean

    {usage}

    Parameters
    ----------
    {common_parameters}

    method : str or callable, optional (default: 'auto')
        The available methods are::

            'auto'       # Use loess if (n<1000), glm otherwise
            'lm', 'ols'  # Linear Model
            'wls'        # Weighted Linear Model
            'rlm'        # Robust Linear Model
            'glm'        # Generalized linear Model
            'gls'        # Generalized Least Squares
            'lowess'     # Locally Weighted Regression (simple)
            'loess'      # Locally Weighted Regression
            'mavg'       # Moving Average
            'gpr'        # Gaussian Process Regressor

        If a `callable` is passed, it must have the signature::

            def my_smoother(data, xseq, **params):
                # * data - has the x and y values for the model
                # * xseq - x values to be predicted
                # * params - stat parameters
                #
                # It must return a new dataframe. Below is the
                # template used internally by Plotnine

                # Input data into the model
                x, y = data['x'], data['y']

                # Create and fit a model
                model = Model(x, y)
                results = Model.fit()

                # Create output data by getting predictions on
                # the xseq values
                data = pd.DataFrame({
                    'x': xseq,
                    'y': results.predict(xseq)})

                # Compute confidence intervals, this depends on
                # the model. However, given standard errors and the
                # degrees of freedom we can compute the confidence
                # intervals using the t-distribution.
                #
                # For an alternative, implement confidence interals by
                # the bootstrap method
                if params['se']:
                    from plotnine.utils.smoothers import tdist_ci
                    y = data['y']            # The predicted value
                    df = 123                 # Degrees of freedom
                    stderr = results.stderr  # Standard error
                    level = params['level']  # The parameter value
                    low, high = tdist_ci(y, df, stderr, level)
                    data['se'] = stderr
                    data['ymin'] = low
                    data['ymax'] = high

                return data

        For *loess* smoothing you must install the `scikit-misc` package.
        You can install it using with ``pip install scikit-misc`` or
        ``pip install plotnine[all]``.
    formula : formula_like
        An object that can be used to construct a patsy design matrix.
        This is usually a string. You can only use a formula if ``method``
        is one of *lm*, *ols*, *wls*, *glm*, *rlm* or *gls*, and in the
        :ref:`formula <patsy:formulas>` you may refer to the ``x`` and
        ``y`` aesthetic variables.
    se : bool (default: True)
        If :py:`True` draw confidence interval around the smooth line.
    n : int (default: 80)
        Number of points to evaluate the smoother at. Some smoothers
        like *mavg* do not support this.
    fullrange : bool (default: False)
        If :py:`True` the fit will span the full range of the plot.
    level : float (default: 0.95)
        Level of confidence to use if :py:`se=True`.
    span : float (default: 2/3.)
        Controls the amount of smoothing for the *loess* smoother.
        Larger number means more smoothing. It should be in the
        ``(0, 1)`` range.
    method_args : dict (default: {})
        Additional arguments passed on to the modelling method.

    See Also
    --------
    statsmodels.regression.linear_model.OLS
    statsmodels.regression.linear_model.WLS
    statsmodels.robust.robust_linear_model.RLM
    statsmodels.genmod.generalized_linear_model.GLM
    statsmodels.regression.linear_model.GLS
    statsmodels.nonparametric.smoothers_lowess.lowess
    skmisc.loess.loess
    pandas.DataFrame.rolling
    sklearn.gaussian_process.GaussianProcessRegressor

    Notes
    -----
    :class:`~plotnine.geoms.geom_smooth` and :class:`.stat_smooth` are
    effectively aliases, they both use the same arguments.
    Use :class:`~plotnine.geoms.geom_smooth` unless
    you want to display the results with a non-standard geom.
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

         'se'    # Standard error of points in bin
         'ymin'  # Lower confidence limit
         'ymax'  # Upper confidence limit

    Calculated aesthetics are accessed using the `after_stat` function.
    e.g. :py:`after_stat('se')`.
    """

    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = {
        "geom": "smooth",
        "position": "identity",
        "na_rm": False,
        "method": "auto",
        "se": True,
        "n": 80,
        "formula": None,
        "fullrange": False,
        "level": 0.95,
        "span": 0.75,
        "method_args": {},
    }
    CREATES = {"se", "ymin", "ymax"}

    def setup_data(self, data):
        """
        Overide to modify data before compute_layer is called
        """
        data = data[np.isfinite(data["x"]) & np.isfinite(data["y"])]
        return data

    def setup_params(self, data):
        params = self.params.copy()
        # Use loess/lowess for small datasets
        # and glm for large
        if params["method"] == "auto":
            max_group = data["group"].value_counts().max()
            if max_group < 1000:
                try:
                    from skmisc.loess import loess  # noqa: F401

                    params["method"] = "loess"
                except ImportError:
                    params["method"] = "lowess"
            else:
                params["method"] = "glm"

        if params["method"] == "mavg":
            if "window" not in params["method_args"]:
                window = len(data) // 10
                warnings.warn(
                    "No 'window' specified in the method_args. "
                    f"Using window = {window}. "
                    "The same window is used for all groups or "
                    "facets",
                    PlotnineWarning,
                )
                params["method_args"]["window"] = window

        if params["formula"]:
            allowed = {"lm", "ols", "wls", "glm", "rlm", "gls"}
            if params["method"] not in allowed:
                raise ValueError(
                    "You can only use a formula with `method` is "
                    f"one of {allowed}"
                )
            params["enviroment"] = self.environment

        return params

    @classmethod
    def compute_group(cls, data, scales, **params):
        data = data.sort_values("x")
        n = params["n"]

        x_unique = data["x"].unique()

        if len(x_unique) < 2:
            warnings.warn(
                "Smoothing requires 2 or more points. Got "
                f"{len(x_unique)}. Not enough points for smoothing. "
                "If this message a surprise, make sure the column "
                "mapped to the x aesthetic has the right dtype.",
                PlotnineWarning,
            )
            # Not enough data to fit
            return pd.DataFrame()

        if data["x"].dtype.kind == "i":
            if params["fullrange"]:
                xseq = scales.x.dimension()
            else:
                xseq = np.sort(x_unique)
        else:
            if params["fullrange"]:
                rangee = scales.x.dimension()
            else:
                rangee = [data["x"].min(), data["x"].max()]
            xseq = np.linspace(rangee[0], rangee[1], n)

        return predictdf(data, xseq, **params)



import pandas as pd

from plotnine import aes, after_stat, ggplot, stat_ecdf

df = pd.DataFrame({"x": range(10)})
p = ggplot(df, aes("x")) + stat_ecdf(size=2)


def test_ecdf():
    p = ggplot(df, aes("x")) + stat_ecdf(size=2)

    assert p == "ecdf"


def test_computed_y_column():
    p = (
        ggplot(df, aes("x"))
        + stat_ecdf(size=2)
        # Should be able to used computed y column & create a
        # new mapped column also called y
        + stat_ecdf(aes(y=after_stat("y-0.2")), size=2, color="blue")
    )
    assert p == "computed_y_column"



import numpy as np
import pandas as pd

from plotnine import (
    aes,
    after_stat,
    coord_flip,
    facet_wrap,
    geom_area,
    geom_line,
    geom_ribbon,
    ggplot,
    scale_x_continuous,
)

n = 4  # No. of ribbions in a vertical stack
m = 100  # Points
width = 2 * np.pi  # width of each ribbon
x = np.linspace(0, width, m)

df = pd.DataFrame(
    {
        "x": np.tile(x, n),
        "ymin": np.hstack([np.sin(x) + 2 * i for i in range(n)]),
        "ymax": np.hstack([np.sin(x) + 2 * i + 1 for i in range(n)]),
        "z": np.repeat(range(n), m),
    }
)


def test_ribbon_aesthetics():
    p = (
        ggplot(df, aes("x", ymin="ymin", ymax="ymax", group="factor(z)"))
        + geom_ribbon()
        + geom_ribbon(aes("x+width", alpha="z"))
        + geom_ribbon(
            aes("x+2*width", linetype="factor(z)"),
            color="black",
            fill=None,
            size=2,
        )
        + geom_ribbon(aes("x+3*width", color="z"), fill=None, size=2)
        + geom_ribbon(aes("x+4*width", fill="factor(z)"))
        + geom_ribbon(aes("x+5*width", size="z"), color="black", fill=None)
        + scale_x_continuous(
            breaks=[i * 2 * np.pi for i in range(7)],
            labels=["0"] + [rf"${2*i}\pi$" for i in range(1, 7)],
        )
    )

    assert p == "ribbon_aesthetics"


def test_area_aesthetics():
    p = (
        ggplot(df, aes("x", "ymax+2", group="factor(z)"))
        + geom_area()
        + geom_area(aes("x+width", alpha="z"))
        + geom_area(
            aes("x+2*width", linetype="factor(z)"),
            color="black",
            fill=None,
            size=2,
        )
        + geom_area(aes("x+3*width", color="z"), fill=None, size=2)
        + geom_area(aes("x+4*width", fill="factor(z)"))
        + geom_area(aes("x+5*width", size="z"), color="black", fill=None)
        + scale_x_continuous(
            breaks=[i * 2 * np.pi for i in range(7)],
            labels=["0"] + [rf"${2*i}\pi$" for i in range(1, 7)],
        )
    )

    assert p == "area_aesthetics"


def test_ribbon_facetting():
    p = (
        ggplot(df, aes("x", ymin="ymin", ymax="ymax", fill="factor(z)"))
        + geom_ribbon()
        + facet_wrap("~ z")
    )

    assert p == "ribbon_facetting"


def test_ribbon_coord_flip():
    df = pd.DataFrame({"x": [0, 1, 2, 3, 4, 5], "y": [0, 3, 5, 5, 3, 0]})

    p = (
        ggplot(df, aes("x"))
        + geom_ribbon(aes(ymax="y"), ymin=0)
        + coord_flip()
    )

    assert p == "ribbon_coord_flip"


def test_ribbon_where():
    m = 3
    n = 100
    values = np.linspace(0, 2 * m * np.pi, n)
    df = pd.DataFrame({"x": range(n), "sin": np.sin(values)})

    p = (
        ggplot(df, aes("x", "sin"))
        + geom_ribbon(
            aes(ymin=0, ymax="sin", where="sin>0"), fill="blue", alpha=0.2
        )
        + geom_ribbon(
            aes(ymin=0, ymax="sin", where="sin<0"), fill="red", alpha=0.2
        )
        + geom_line()
    )
    assert p == "ribbon_where"


class TestOutlineType:
    x = np.arange(10)
    d = 5
    df = pd.DataFrame(
        {"x": x, "y1": x, "y2": x + 2 * d, "y3": x + 4 * d, "y4": x + 6 * d}
    )

    p = (
        ggplot(df, aes("x", ymax=after_stat("ymin + d")))
        + geom_ribbon(
            aes(ymin="y1"),
            size=1,
            fill="bisque",
            color="orange",
            outline_type="upper",
        )
        + geom_ribbon(
            aes(ymin="y2"),
            size=1,
            fill="khaki",
            color="darkkhaki",
            outline_type="lower",
        )
        + geom_ribbon(
            aes(ymin="y3"),
            size=1,
            fill="plum",
            color="purple",
            outline_type="both",
        )
        + geom_ribbon(
            aes(ymin="y4"),
            size=1,
            fill="lightblue",
            color="cadetblue",
            outline_type="full",
        )
    )

    def test_ribbon_outline_type(self):
        assert self.p == "ribbon_outline_type"

    def test_ribbon_outline_type_coord_flip(self):
        assert self.p + coord_flip() == "ribbon_outline_type_coord_flip"



from __future__ import annotations

import typing
from copy import deepcopy
from warnings import warn

import numpy as np
import pandas as pd

from ..exceptions import PlotnineError, PlotnineWarning
from ..utils import Registry, is_string
from .guide import guide as guide_class

if typing.TYPE_CHECKING:
    from typing import Literal

    from plotnine.typing import TupleFloat2

# Terminology
# -----------
# - A guide is either a legend or colorbar.
#
# - A guide definition (gdef) is an instantiated guide as it
#   is used in the process of creating the legend
#
# - The guides class holds all guides that will appear in the
#   plot
#
# - A guide box is a fully drawn out guide.
#   It is of subclass matplotlib.offsetbox.Offsetbox


class guides(dict):
    """
    Guides for each scale

    Used to assign a particular guide to an aesthetic(s).

    Parameters
    ----------
    kwargs : dict
        aesthetic - guide pairings. e.g
        ``color=guide_colorbar()``
    """

    # Determined from the theme when the guides are
    # getting built
    position: Literal["left", "right", "top", "bottom"] | TupleFloat2
    box_direction: Literal["horizontal", "vertical", "auto"]
    box_align: Literal["left", "right", "top", "bottom", "center", "auto"]
    box_margin: int
    spacing: float

    def __init__(self, **kwargs):
        aes_names = {
            "alpha",
            "color",
            "fill",
            "linetype",
            "shape",
            "size",
            "stroke",
        }
        if "colour" in kwargs:
            kwargs["color"] = kwargs.pop("colour")

        dict.__init__(
            self, ((ae, kwargs[ae]) for ae in kwargs if ae in aes_names)
        )

    def __radd__(self, gg):
        """
        Add guides to the plot

        Parameters
        ----------
        gg : ggplot
            ggplot object being created

        Returns
        -------
        out : gglot
            ggplot object with guides. If *inplace*
            is **False** this is a copy of the original
            ggplot object.
        """
        new_guides = {}
        for k in self:
            new_guides[k] = deepcopy(self[k])
        gg.guides.update(new_guides)
        return gg

    def build(self, plot):
        """
        Build the guides

        Parameters
        ----------
        plot : ggplot
            ggplot object being drawn

        Returns
        -------
        box : matplotlib.offsetbox.Offsetbox | None
            A box that contains all the guides for the plot.
            If there are no guides, **None** is returned.
        """
        _property = plot.theme.themeables.property
        self.box_direction = _property("legend_box")
        self.position = _property("legend_position")
        self.box_align = _property("legend_box_just")
        self.box_margin = _property("legend_box_margin")
        self.spacing = _property("legend_spacing")

        if self.position == "none":
            return  # No Legend

        # Direction
        if self.box_direction == "auto":
            if self.position in ("right", "left"):
                self.box_direction = "vertical"
            else:
                self.box_direction = "horizontal"

        # Justification of legend boxes
        if self.box_align == "auto":
            if self.position in ("right", "left"):
                self.box_align = "left"
            else:
                self.box_align = "right"

        gdefs = self.train(plot)
        if not gdefs:
            return

        gdefs = self.merge(gdefs)
        gdefs = self.create_geoms(gdefs, plot)

        if not gdefs:
            return

        gboxes = self.draw(gdefs, plot.theme)
        bigbox = self.assemble(gboxes, gdefs, plot.theme)
        return bigbox

    def train(self, plot):
        """
        Compute all the required guides

        Parameters
        ----------
        plot : ggplot
            ggplot object

        Returns
        -------
        gdefs : list
            Guides for the plots
        """
        gdefs = []

        for scale in plot.scales:
            for output in scale.aesthetics:
                # The guide for aesthetic 'xxx' is stored
                # in plot.guides['xxx']. The priority for
                # the guides depends on how they are created
                # 1. ... + guides(xxx=guide_blah())
                # 2. ... + scale_xxx(guide=guide_blah())
                # 3. default(either guide_legend or guide_colorbar
                #            depending on the scale type)
                # output = scale.aesthetics[0]
                guide = self.get(output, scale.guide)

                if guide is None or guide is False:
                    continue

                # check the validity of guide.
                # if guide is character, then find the guide object
                guide = self.validate(guide)
                # check the consistency of the guide and scale.
                if (
                    "any" not in guide.available_aes
                    and scale.aesthetics[0] not in guide.available_aes
                ):
                    raise PlotnineError(
                        f"{guide.__class__.__name__} cannot be used for "
                        f"{scale.aesthetics}"
                    )

                # title
                if not hasattr(guide, "title"):
                    if scale.name:
                        guide.title = scale.name
                    else:
                        guide.title = getattr(plot.labels, output)
                        if guide.title is None:
                            warn(
                                f"Cannot generate legend for the {output!r} "
                                "aesthetic. Make sure you have mapped a "
                                "variable to it",
                                PlotnineWarning,
                            )

                # each guide object trains scale within the object,
                # so Guides (i.e., the container of guides)
                # need not to know about them
                guide = guide.train(scale, output)

                if guide is not None:
                    gdefs.append(guide)

        return gdefs

    def validate(self, guide):
        """
        Validate guide object
        """
        if is_string(guide):
            guide = Registry[f"guide_{guide}"]()

        if not isinstance(guide, guide_class):
            raise PlotnineError(f"Unknown guide: {guide}")
        return guide

    def merge(self, gdefs):
        """
        Merge overlapped guides

        For example::

            from plotnine import *
            gg = ggplot(mtcars, aes(y='wt', x='mpg', colour='factor(cyl)'))
            gg = gg + stat_smooth(aes(fill='factor(cyl)'), method='lm')
            gg = gg + geom_point()
            gg

        This would create two guides with the same hash
        """
        # group guide definitions by hash, and
        # reduce each group to a single guide
        # using the guide.merge method
        df = pd.DataFrame({"gdef": gdefs, "hash": [g.hash for g in gdefs]})
        grouped = df.groupby("hash", sort=False)
        gdefs = []
        for name, group in grouped:
            # merge
            gdef = group["gdef"].iloc[0]
            for g in group["gdef"].iloc[1:]:
                gdef = gdef.merge(g)
            gdefs.append(gdef)
        return gdefs

    def create_geoms(self, gdefs, plot):
        """
        Add geoms to the guide definitions
        """
        new_gdefs = []
        for gdef in gdefs:
            gdef = gdef.create_geoms(plot)
            if gdef:
                new_gdefs.append(gdef)

        return new_gdefs

    def draw(self, gdefs, theme):
        """
        Draw out each guide definition

        Parameters
        ----------
        gdefs : list of guide_legend|guide_colorbar
            guide definitions
        theme : theme
            Plot theme

        Returns
        -------
        out : list of matplotlib.offsetbox.Offsetbox
            A drawing of each legend
        """
        for g in gdefs:
            g._set_defaults(theme)
        return [g.draw() for g in gdefs]

    def assemble(self, gboxes, gdefs, theme):
        """
        Put together all the guide boxes

        Parameters
        ----------
        gboxes : list
            List of :class:`~matplotlib.offsetbox.Offsetbox`,
            where each item is a legend for a single aesthetic.
        gdefs : list of guide_legend|guide_colorbar
            guide definitions
        theme : theme
            Plot theme

        Returns
        -------
        box : OffsetBox
            A box than can be placed onto a plot
        """
        from matplotlib.offsetbox import HPacker, VPacker

        # place the guides according to the guide.order
        # 0 do not sort
        # 1-99 sort
        for gdef in gdefs:
            if gdef.order == 0:
                gdef.order = 100
            elif not 0 <= gdef.order <= 99:
                raise PlotnineError(
                    "'order' for a guide should be " "between 0 and 99"
                )
        orders = [gdef.order for gdef in gdefs]
        idx = np.argsort(orders)
        gboxes = [gboxes[i] for i in idx]

        # direction when more than legend
        if self.box_direction == "vertical":
            packer = VPacker
        elif self.box_direction == "horizontal":
            packer = HPacker
        else:
            raise PlotnineError(
                "'legend_box' should be either " "'vertical' or 'horizontal'"
            )

        box = packer(
            children=gboxes,
            align=self.box_align,
            pad=self.box_margin,
            sep=self.spacing,
        )
        return box



"""
Scales
"""
# limits
from .limits import expand_limits, lims, xlim, ylim

# alpha
from .scale_alpha import (
    scale_alpha,
    scale_alpha_continuous,  # pyright: ignore
    scale_alpha_datetime,
    scale_alpha_discrete,
    scale_alpha_ordinal,
)

# pyright: reportGeneralTypeIssues=true
# fill
# color
from .scale_color import (
    scale_color_brewer,
    scale_color_cmap,
    scale_color_cmap_d,
    scale_color_continuous,  # pyright: ignore
    scale_color_datetime,
    scale_color_desaturate,
    scale_color_discrete,  # pyright: ignore
    scale_color_distiller,
    scale_color_gradient,
    scale_color_gradient2,
    scale_color_gradientn,
    scale_color_gray,  # pyright: ignore
    scale_color_grey,
    scale_color_hue,
    scale_color_ordinal,  # pyright: ignore
    scale_colour_brewer,  # pyright: ignore
    scale_colour_cmap,  # pyright: ignore
    scale_colour_cmap_d,  # pyright: ignore
    scale_colour_continuous,  # pyright: ignore
    scale_colour_datetime,  # pyright: ignore
    scale_colour_desaturate,  # pyright: ignore
    scale_colour_discrete,  # pyright: ignore
    scale_colour_distiller,  # pyright: ignore
    scale_colour_gradient,  # pyright: ignore
    scale_colour_gradient2,  # pyright: ignore
    scale_colour_gradientn,  # pyright: ignore
    scale_colour_gray,  # pyright: ignore
    scale_colour_grey,  # pyright: ignore
    scale_colour_hue,  # pyright: ignore
    scale_colour_ordinal,  # pyright: ignore
    scale_fill_brewer,
    scale_fill_cmap,
    scale_fill_cmap_d,
    scale_fill_continuous,  # pyright: ignore
    scale_fill_datetime,
    scale_fill_desaturate,
    scale_fill_discrete,  # pyright: ignore
    scale_fill_distiller,
    scale_fill_gradient,
    scale_fill_gradient2,
    scale_fill_gradientn,
    scale_fill_gray,  # pyright: ignore
    scale_fill_grey,
    scale_fill_hue,
    scale_fill_ordinal,  # pyright: ignore
)

# identity
from .scale_identity import (
    scale_alpha_identity,
    scale_color_identity,
    scale_colour_identity,  # pyright: ignore
    scale_fill_identity,
    scale_linetype_identity,
    scale_shape_identity,
    scale_size_identity,
)

# linetype
from .scale_linetype import (
    scale_linetype,
    scale_linetype_continuous,
    scale_linetype_discrete,  # pyright: ignore
)

# manual
from .scale_manual import (
    scale_alpha_manual,
    scale_color_manual,
    scale_colour_manual,  # pyright: ignore
    scale_fill_manual,
    scale_linetype_manual,
    scale_shape_manual,
    scale_size_manual,
)

# shape
from .scale_shape import (
    scale_shape,
    scale_shape_continuous,
    scale_shape_discrete,  # pyright: ignore
)

# size
from .scale_size import (
    scale_size,  # pyright: ignore
    scale_size_area,
    scale_size_continuous,
    scale_size_datetime,
    scale_size_discrete,
    scale_size_ordinal,
    scale_size_radius,
)

# stroke
from .scale_stroke import (
    scale_stroke,  # pyright: ignore
    scale_stroke_continuous,
    scale_stroke_discrete,
)

# xy position and transforms
from .scale_xy import (
    scale_x_continuous,
    scale_x_date,  # pyright: ignore
    scale_x_datetime,
    scale_x_discrete,
    scale_x_log10,
    scale_x_reverse,
    scale_x_sqrt,
    scale_x_timedelta,
    scale_y_continuous,
    scale_y_date,  # pyright: ignore
    scale_y_datetime,
    scale_y_discrete,
    scale_y_log10,
    scale_y_reverse,
    scale_y_sqrt,
    scale_y_timedelta,
)

__all__ = [
    # color
    "scale_color_brewer",
    "scale_colour_brewer",
    "scale_color_cmap",
    "scale_colour_cmap",
    "scale_color_cmap_d",
    "scale_colour_cmap_d",
    "scale_color_ordinal",
    "scale_colour_ordinal",
    "scale_color_continuous",
    "scale_colour_continuous",
    "scale_color_discrete",
    "scale_colour_discrete",
    "scale_color_distiller",
    "scale_colour_distiller",
    "scale_color_desaturate",
    "scale_colour_desaturate",
    "scale_color_gradient",
    "scale_colour_gradient",
    "scale_color_gradient2",
    "scale_colour_gradient2",
    "scale_color_gradientn",
    "scale_colour_gradientn",
    "scale_color_grey",
    "scale_colour_grey",
    "scale_color_gray",
    "scale_colour_gray",
    "scale_color_hue",
    "scale_colour_hue",
    "scale_color_datetime",
    "scale_colour_datetime",
    # fill
    "scale_fill_brewer",
    "scale_fill_cmap",
    "scale_fill_cmap_d",
    "scale_fill_ordinal",
    "scale_fill_continuous",
    "scale_fill_desaturate",
    "scale_fill_discrete",
    "scale_fill_distiller",
    "scale_fill_gradient",
    "scale_fill_gradient2",
    "scale_fill_gradientn",
    "scale_fill_grey",
    "scale_fill_gray",
    "scale_fill_hue",
    "scale_fill_datetime",
    # alpha
    "scale_alpha",
    "scale_alpha_discrete",
    "scale_alpha_ordinal",
    "scale_alpha_continuous",
    "scale_alpha_datetime",
    # linetype
    "scale_linetype",
    "scale_linetype_discrete",
    "scale_linetype_continuous",
    # shape
    "scale_shape",
    "scale_shape_discrete",
    "scale_shape_continuous",
    # size
    "scale_size",
    "scale_size_area",
    "scale_size_discrete",
    "scale_size_continuous",
    "scale_size_ordinal",
    "scale_size_radius",
    "scale_size_datetime",
    # stroke
    "scale_stroke",
    "scale_stroke_continuous",
    "scale_stroke_discrete",
    # identity
    "scale_alpha_identity",
    "scale_color_identity",
    "scale_colour_identity",
    "scale_fill_identity",
    "scale_linetype_identity",
    "scale_shape_identity",
    "scale_size_identity",
    # manual
    "scale_color_manual",
    "scale_colour_manual",
    "scale_fill_manual",
    "scale_shape_manual",
    "scale_linetype_manual",
    "scale_alpha_manual",
    "scale_size_manual",
    # xy position and transforms
    "scale_x_continuous",
    "scale_x_date",
    "scale_x_datetime",
    "scale_x_discrete",
    "scale_x_log10",
    "scale_x_reverse",
    "scale_x_sqrt",
    "scale_x_timedelta",
    "scale_y_continuous",
    "scale_y_date",
    "scale_y_datetime",
    "scale_y_discrete",
    "scale_y_log10",
    "scale_y_reverse",
    "scale_y_sqrt",
    "scale_y_timedelta",
    # limits
    "xlim",
    "ylim",
    "lims",
    "expand_limits",
]



import pandas as pd
import pytest

from plotnine import aes, geom_abline, geom_point, ggplot
from plotnine.exceptions import PlotnineWarning

df = pd.DataFrame(
    {
        "slope": [1, 1],
        "intercept": [1, -1],
        "x": [-1, 1],
        "y": [-1, 1],
        "z": range(2),
    }
)


def test_aesthetics():
    p = (
        ggplot(df, aes("x", "y"))
        + geom_point()
        + geom_abline(aes(slope="slope", intercept="intercept"), size=2)
        + geom_abline(
            aes(slope="slope", intercept="intercept+.1", alpha="z"), size=2
        )
        + geom_abline(
            aes(slope="slope", intercept="intercept+.2", linetype="factor(z)"),
            size=2,
        )
        + geom_abline(
            aes(slope="slope", intercept="intercept+.3", color="factor(z)"),
            size=2,
        )
        + geom_abline(aes(slope="slope", intercept="intercept+.4", size="z"))
    )

    assert p == "aesthetics"


def test_aes_inheritance():
    # A default line (intercept = 0, slope = 1)
    p = (
        ggplot(
            df,
            aes(
                "x",
                "y",
                color="factor(z)",
                slope="slope",
                intercept="intercept",
            ),
        )
        + geom_point(size=10, show_legend=False)
        + geom_abline(size=2)
    )

    assert p == "aes_inheritance"


def test_aes_overwrite():
    with pytest.warns(PlotnineWarning):
        geom_abline(aes(intercept="y"), intercept=2)



from __future__ import annotations

import typing

from ..doctools import document
from ..utils import to_rgba
from .geom import geom
from .geom_line import geom_line, geom_path
from .geom_ribbon import geom_ribbon

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord, DrawingArea, Layer


@document
class geom_smooth(geom):
    """
    A smoothed conditional mean

    {usage}

    Parameters
    ----------
    {common_parameters}
    legend_fill_ratio : float (default: 0.5)
        How much (vertically) of the legend box should be filled by
        the color that indicates the confidence intervals. Should be
        in the range [0, 1].
    """

    DEFAULT_AES = {
        "alpha": 0.4,
        "color": "black",
        "fill": "#999999",
        "linetype": "solid",
        "size": 1,
        "ymin": None,
        "ymax": None,
    }
    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = {
        "stat": "smooth",
        "position": "identity",
        "na_rm": False,
        "legend_fill_ratio": 0.5,
    }

    def use_defaults(
        self, data: pd.DataFrame, aes_modifiers: dict[str, Any]
    ) -> pd.DataFrame:
        has_ribbon = "ymin" in data and "ymax" in data
        data = super().use_defaults(data, aes_modifiers)

        # When there is no ribbon, the default values for 'ymin'
        # and 'ymax' are None (not numeric). So we remove them
        # prevent any computations that may use them without checking.
        if not has_ribbon:
            del data["ymin"]
            del data["ymax"]
        return data

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        return data.sort_values(["PANEL", "group", "x"])

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        has_ribbon = "ymin" in data and "ymax" in data
        if has_ribbon:
            data2 = data.copy()
            data2["color"] = "none"
            params2 = params.copy()
            params2["outline_type"] = "full"
            geom_ribbon.draw_group(data2, panel_params, coord, ax, **params2)

        data["alpha"] = 1
        geom_line.draw_group(data, panel_params, coord, ax, **params)

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw letter 'a' in the box

        Parameters
        ----------
        data : dataframe
            Legend data
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        from matplotlib.patches import Rectangle

        try:
            has_se = lyr.stat.params["se"]
        except KeyError:
            has_se = False

        if has_se:
            fill = to_rgba(data["fill"], data["alpha"])
            r = lyr.geom.params["legend_fill_ratio"]
            bg = Rectangle(
                (0, (1 - r) * da.height / 2),
                width=da.width,
                height=r * da.height,
                facecolor=fill,
                linewidth=0,
            )
            da.add_artist(bg)

        data["alpha"] = 1
        return geom_path.draw_legend(data, da, lyr)



import warnings
from pathlib import Path

import matplotlib.pyplot as plt
import pytest

from plotnine import aes, geom_text, ggplot, ggtitle, theme
from plotnine.data import mtcars
from plotnine.exceptions import PlotnineError, PlotnineWarning
from plotnine.ggplot import save_as_pdf_pages


def p(N=3):
    """Return *N* distinct plot objects."""
    template = ggplot(mtcars, aes(x="wt", y="mpg", label="name")) + geom_text()
    for i in range(1, N + 1):
        yield template + ggtitle("%d of %d" % (i, N))


def sequential_filenames():
    """
    Generate filenames for the tests
    """
    for i in range(100):
        yield Path(f"filename-{i}.png")


filename_gen = sequential_filenames()


def assert_exist_and_clean(filename, msg=None):
    if not msg:
        msg = f"File {filename} does not exist"
    assert filename.exists(), msg
    filename.unlink()


class TestArguments:
    def test_default_filename(self):
        plots = list(p())
        save_as_pdf_pages(plots, verbose=False)
        fn = plots[0]._save_filename("pdf")
        assert_exist_and_clean(fn, "default filename")

    def test_save_method(self):
        fn = next(filename_gen)
        with pytest.warns(UserWarning) as record:
            save_as_pdf_pages(p(), fn)

        assert_exist_and_clean(fn, "save method")

        res = ("filename" in str(item.message).lower() for item in record)
        assert any(res)

        # verbose
        fn = next(filename_gen)
        with warnings.catch_warnings(record=True) as record:
            save_as_pdf_pages(p(), fn, verbose=False)
            assert_exist_and_clean(fn, "save method")
            assert not record, "Issued an unexpected warning"

        res = ("filename" in str(item.message).lower() for item in record)
        assert not any(res)

    def test_filename_plot_path(self):
        fn = next(filename_gen)
        with pytest.warns(PlotnineWarning):
            save_as_pdf_pages(p(), fn, path=".")
        assert_exist_and_clean(fn, "fn, plot and path")

    @pytest.mark.skip(
        "Results of this test can only be confirmed by"
        "inspecting the generated PDF."
    )
    def test_height_width(self):
        plots = []
        for i, plot in enumerate(p()):
            plots.append(plot + theme(figure_size=(8 + i, 6 + i)))
        fn = next(filename_gen)
        with pytest.warns(PlotnineWarning):
            save_as_pdf_pages(plots, fn)
        # assert False, "Check %s" % fn  # Uncomment to check


class TestExceptions:
    def test_plot_exception(self):
        # Force an error in drawing
        fn = next(filename_gen)
        plots = list(p())
        plots[0] += aes(color="unknown")
        with pytest.raises(PlotnineError):
            save_as_pdf_pages(plots, fn, verbose=False)

        assert_exist_and_clean(fn, "Plot exception")


# This should be the last function in the file since it can catch
# "leakages" due to the tests in this test module.
def test_save_as_pdf_pages_closes_plots():
    assert plt.get_fignums() == [], "There are unsaved test plots"
    fn = next(filename_gen)
    with pytest.warns(PlotnineWarning):
        save_as_pdf_pages(p(), fn)
    assert_exist_and_clean(fn, "exist")
    assert plt.get_fignums() == [], "ggplot.save did not close the plot"



from __future__ import annotations

import typing

import numpy as np

from ..doctools import document
from ..utils import SIZE_FACTOR, to_rgba
from .geom import geom

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord, DrawingArea, Layer


@document
class geom_polygon(geom):
    """
    Polygon, a filled path

    {usage}

    Parameters
    ----------
    {common_parameters}

    Notes
    -----
    All paths in the same ``group`` aesthetic value make up a polygon.
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": None,
        "fill": "#333333",
        "linetype": "solid",
        "size": 0.5,
    }
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
    }
    REQUIRED_AES = {"x", "y"}

    def handle_na(self, data: pd.DataFrame) -> pd.DataFrame:
        return data

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot all groups
        """
        self.draw_group(data, panel_params, coord, ax, **params)

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        from matplotlib.collections import PolyCollection

        data = coord.transform(data, panel_params, munch=True)
        data["size"] *= SIZE_FACTOR

        # Each group is a polygon with a single facecolor
        # with potentially an edgecolor for every edge.
        verts = []
        facecolor = []
        edgecolor = []
        linestyle = []
        linewidth = []

        # Some stats may order the data in ways that prevent
        # objects from occluding other objects. We do not want
        # to undo that order.
        grouper = data.groupby("group", sort=False)
        for group, df in grouper:
            fill = to_rgba(df["fill"].iloc[0], df["alpha"].iloc[0])
            verts.append(tuple(zip(df["x"], df["y"])))
            facecolor.append("none" if fill is None else fill)
            edgecolor.append(df["color"].iloc[0] or "none")
            linestyle.append(df["linetype"].iloc[0])
            linewidth.append(df["size"].iloc[0])

        col = PolyCollection(
            verts,
            facecolors=facecolor,
            edgecolors=edgecolor,
            linestyles=linestyle,
            linewidths=linewidth,
            zorder=params["zorder"],
            rasterized=params["raster"],
        )

        ax.add_collection(col)

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw a rectangle in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        from matplotlib.patches import Rectangle

        data["size"] *= SIZE_FACTOR
        # We take into account that the linewidth
        # bestrides the boundary of the rectangle
        linewidth = np.min([data["size"], da.width / 4, da.height / 4])

        if data["color"] is None:
            linewidth = 0

        facecolor = to_rgba(data["fill"], data["alpha"])
        if facecolor is None:
            facecolor = "none"

        rect = Rectangle(
            (0 + linewidth / 2, 0 + linewidth / 2),
            width=da.width - linewidth,
            height=da.height - linewidth,
            linewidth=linewidth,
            linestyle=data["linetype"],
            facecolor=facecolor,
            edgecolor=data["color"],
            capstyle="projecting",
        )
        da.add_artist(rect)
        return da



from __future__ import annotations

import typing

from matplotlib.patches import FancyBboxPatch
from matplotlib.text import _get_textbox
from matplotlib.transforms import Affine2D

if typing.TYPE_CHECKING:
    from matplotlib.backend_bases import RendererBase

    from plotnine.typing import StripPosition

    from .text import SText


# We subclass because we want to learn the size and location
# of the box when the layout manager is running.
# With MPLs default text & boxpatch, the patch gets its
# dimension information at draw time.


class SFancyBboxPatch(FancyBboxPatch):
    """
    Strip text background box
    """

    # The text artists that is wrapped by this box
    text: SText
    position: StripPosition
    _update = False

    def __init__(self, text: SText):
        boxstyle = f"square, pad={text.draw_info.strip_text_margin}"

        super().__init__(
            (0, 0), 1, 1, boxstyle=boxstyle, clip_on=False, zorder=2.2
        )

        self.text = text
        self.position = text.draw_info.position

    def update_position_size(self, renderer: RendererBase):
        """
        Update the location and the size of the bbox.
        """
        if self._update:
            return

        text = self.text
        posx, posy = text.get_transform().transform((text._x, text._y))
        x, y, w, h = _get_textbox(text, renderer)

        self.set_bounds(0.0, 0.0, w, h)
        self.set_transform(
            Affine2D()
            .rotate_deg(text.get_rotation())
            .translate(posx + x, posy + y)
        )
        fontsize_in_pixel = renderer.points_to_pixels(
            text.get_size()  # pyright: ignore
        )
        self.set_mutation_scale(fontsize_in_pixel)
        self._update = True

    def get_window_extent(self, renderer):
        """
        Location & dimensions of the box
        """
        self.update_position_size(renderer)
        return super().get_window_extent(renderer)



from __future__ import annotations

import typing
from copy import copy
from warnings import warn

import numpy as np

from ..exceptions import PlotnineError, PlotnineWarning
from ..mapping.aes import X_AESTHETICS, Y_AESTHETICS
from ..utils import (
    Registry,
    check_required_aesthetics,
    groupby_apply,
    is_string,
)

if typing.TYPE_CHECKING:
    from typing import Any, Optional

    import pandas as pd

    from plotnine.iapi import pos_scales
    from plotnine.typing import Geom, Layout, TransformCol


class position(metaclass=Registry):
    """Base class for all positions"""

    __base__ = True

    REQUIRED_AES: set[str] = set()
    params: dict[str, Any]

    def __init__(self):
        self.params = {}

    def setup_params(self, data: pd.DataFrame) -> dict[str, Any]:
        """
        Verify, modify & return a copy of the params.
        """
        return copy(self.params)

    def setup_data(
        self, data: pd.DataFrame, params: dict[str, Any]
    ) -> pd.DataFrame:
        """
        Verify & return data
        """
        check_required_aesthetics(
            self.REQUIRED_AES, data.columns, self.__class__.__name__
        )
        return data

    @classmethod
    def compute_layer(
        cls, data: pd.DataFrame, params: dict[str, Any], layout: Layout
    ):
        """
        Compute position for the layer in all panels

        Positions can override this function instead of
        `compute_panel` if the position computations are
        independent of the panel. i.e when not colliding
        """

        def fn(pdata: pd.DataFrame) -> pd.DataFrame:
            """
            Compute function helper
            """
            # Given data belonging to a specific panel, grab
            # the corresponding scales and call the method
            # that does the real computation
            if len(pdata) == 0:
                return pdata
            scales = layout.get_scales(pdata["PANEL"].iat[0])
            return cls.compute_panel(pdata, scales, params)

        return groupby_apply(data, "PANEL", fn)

    @classmethod
    def compute_panel(
        cls, data: pd.DataFrame, scales: pos_scales, params: dict[str, Any]
    ) -> pd.DataFrame:
        """
        Positions must override this function

        Notes
        -----
        Make necessary adjustments to the columns in the dataframe.

        Create the position transformation functions and
        use self.transform_position() do the rest.

        See Also
        --------
        position_jitter.compute_panel
        """
        msg = "{} needs to implement this method"
        raise NotImplementedError(msg.format(cls.__name__))

    @staticmethod
    def transform_position(
        data,
        trans_x: Optional[TransformCol] = None,
        trans_y: Optional[TransformCol] = None,
    ) -> pd.DataFrame:
        """
        Transform all the variables that map onto the x and y scales.

        Parameters
        ----------
        data : dataframe
            Data to transform
        trans_x : function
            Transforms x scale mappings
            Takes one argument, either a scalar or an array-type
        trans_y : function
            Transforms y scale mappings
            Takes one argument, either a scalar or an array-type
        """
        if trans_x:
            xs = [name for name in data.columns if name in X_AESTHETICS]
            data[xs] = data[xs].apply(trans_x)

        if trans_y:
            ys = [name for name in data.columns if name in Y_AESTHETICS]
            data[ys] = data[ys].apply(trans_y)

        return data

    @staticmethod
    def from_geom(geom: Geom) -> position:
        """
        Create and return a position object for the geom

        Parameters
        ----------
        geom : geom
            An instantiated geom object.

        Returns
        -------
        out : position
            A position object

        Raises
        ------
        PlotnineError
            If unable to create a `position`.
        """
        name = geom.params["position"]
        if issubclass(type(name), position):
            return name

        if isinstance(name, type) and issubclass(name, position):
            klass = name
        elif is_string(name):
            if not name.startswith("position_"):
                name = f"position_{name}"
            klass = Registry[name]
        else:
            raise PlotnineError(f"Unknown position of type {type(name)}")

        return klass()

    @staticmethod
    def strategy(data: pd.DataFrame, params: dict[str, Any]) -> pd.DataFrame:
        """
        Calculate boundaries of geometry object
        """
        return data

    @classmethod
    def _collide_setup(cls, data, params):
        xminmax = ["xmin", "xmax"]
        width = params.get("width", None)

        # Determine width
        if width is not None:
            # Width set manually
            if not all(col in data.columns for col in xminmax):
                data["xmin"] = data["x"] - width / 2
                data["xmax"] = data["x"] + width / 2
        else:
            if not all(col in data.columns for col in xminmax):
                data["xmin"] = data["x"]
                data["xmax"] = data["x"]

            # Width determined from data, must be floating point constant
            widths = (data["xmax"] - data["xmin"]).drop_duplicates()
            widths = widths[~np.isnan(widths)]
            width = widths.iloc[0]

        return data, width

    @classmethod
    def collide(cls, data, params):
        """
        Calculate boundaries of geometry object

        Uses Strategy
        """
        xminmax = ["xmin", "xmax"]
        data, width = cls._collide_setup(data, params)
        if params.get("width", None) is None:
            params["width"] = width

        # Reorder by x position then on group, relying on stable sort to
        # preserve existing ordering. The default stacking order reverses
        # the group in order to match the legend order.
        if params and "reverse" in params and params["reverse"]:
            idx = data.sort_values(["xmin", "group"], kind="mergesort").index
        else:
            data["-group"] = -data["group"]
            idx = data.sort_values(["xmin", "-group"], kind="mergesort").index
            del data["-group"]

        data = data.loc[idx, :]

        # Check for overlap
        intervals = data[xminmax].drop_duplicates().values.flatten()
        intervals = intervals[~np.isnan(intervals)]

        if len(np.unique(intervals)) > 1 and any(
            np.diff(intervals - intervals.mean()) < -1e-6
        ):
            msg = "{} requires non-overlapping x intervals"
            warn(msg.format(cls.__name__), PlotnineWarning)

        if "ymax" in data:
            data = groupby_apply(data, "xmin", cls.strategy, params)
        elif "y" in data:
            data["ymax"] = data["y"]
            data = groupby_apply(data, "xmin", cls.strategy, params)
            data["y"] = data["ymax"]
        else:
            raise PlotnineError("Neither y nor ymax defined")

        return data

    @classmethod
    def collide2(cls, data, params):
        """
        Calculate boundaries of geometry object

        Uses Strategy
        """
        data, width = cls._collide_setup(data, params)
        if params.get("width", None) is None:
            params["width"] = width

        # Reorder by x position then on group, relying on stable sort to
        # preserve existing ordering. The default stacking order reverses
        # the group in order to match the legend order.
        if params and "reverse" in params and params["reverse"]:
            data["-group"] = -data["group"]
            idx = data.sort_values(["x", "-group"], kind="mergesort").index
            del data["-group"]
        else:
            idx = data.sort_values(["x", "group"], kind="mergesort").index

        data = data.loc[idx, :]
        data.reset_index(inplace=True, drop=True)
        return cls.strategy(data, params)


transform_position = position.transform_position



import numpy as np
import pandas as pd

from ..doctools import document
from .stat import stat


@document
class stat_hull(stat):
    """
    2 Dimensional Convex Hull

    {usage}

    Parameters
    ----------
    {common_parameters}
    qhull_options: str, optional
        Additional options to pass to Qhull.
        See `Qhull <http://www.qhull.org/>`__ documentation
        for details.

    Raises
    ------
    QhullError
        Raised when Qhull encounters an error condition,
        such as geometrical degeneracy when options to resolve are
        not enabled.
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

        'area'  # Area of the convex hull

    """
    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = {
        "geom": "path",
        "position": "identity",
        "na_rm": False,
        "qhull_options": None,
    }
    CREATES = {"area"}

    @classmethod
    def compute_group(cls, data, scales, **params):
        from scipy.spatial import ConvexHull

        hull = ConvexHull(
            data[["x", "y"]], qhull_options=params["qhull_options"]
        )
        idx = np.hstack([hull.vertices, hull.vertices[0]])

        new_data = pd.DataFrame(
            {
                "x": data["x"].iloc[idx].values,
                "y": data["y"].iloc[idx].values,
                "area": hull.area,
            }
        )
        return new_data



from __future__ import annotations

import typing
from contextlib import suppress
from warnings import warn

import numpy as np
import pandas as pd

from .exceptions import PlotnineError, PlotnineWarning
from .facets import facet_grid, facet_null, facet_wrap
from .facets.facet_grid import parse_grid_facets
from .facets.facet_wrap import parse_wrap_facets
from .ggplot import ggplot
from .labels import labs
from .mapping.aes import ALL_AESTHETICS, SCALED_AESTHETICS, aes
from .scales import lims, scale_x_log10, scale_y_log10
from .themes import theme
from .utils import Registry, array_kind

if typing.TYPE_CHECKING:
    from typing import Any, Iterable, Literal

    from plotnine.typing import DataLike, Ggplot, TupleFloat2


def qplot(
    x: str | Iterable[Any] | range | None = None,
    y: str | Iterable[Any] | range | None = None,
    data: DataLike | None = None,
    facets: str = "",
    margins: bool | list[str] = False,
    geom: str | list[str] | tuple[str] = "auto",
    xlim: TupleFloat2 | None = None,
    ylim: TupleFloat2 | None = None,
    log: Literal["x", "y", "xy"] | None = None,
    main: str | None = None,
    xlab: str | None = None,
    ylab: str | None = None,
    asp: float | None = None,
    **kwargs: Any,
) -> Ggplot:
    """
    Quick plot

    Parameters
    ----------
    x : str | array_like
        x aesthetic
    y : str | array_like
        y aesthetic
    data : dataframe
        Data frame to use (optional). If not specified,
        will create one, extracting arrays from the
        current environment.
    geom : str | list
        *geom(s)* to do the drawing. If ``auto``, defaults
        to 'point' if ``x`` and ``y`` are specified or
        'histogram' if only ``x`` is specified.
    facets : str
        Facets
    margins : bool | list[str]
        variable names to compute margins for. True will compute
        all possible margins. Depends on the facetting.
    xlim : tuple
        x-axis limits
    ylim : tuple
        y-axis limits
    log : str in ``{'x', 'y', 'xy'}``
        Which (if any) variables to log transform.
    main : str
        Plot title
    xlab : str
        x-axis label
    ylab : str
        y-axis label
    asp : str | float
        The y/x aspect ratio.
    **kwargs : dict
        Arguments passed on to the geom.

    Returns
    -------
    p : ggplot
        ggplot object
    """
    from patsy.eval import EvalEnvironment

    # Extract all recognizable aesthetic mappings from the parameters
    # String values e.g  "I('red')", "I(4)" are not treated as mappings
    environment = EvalEnvironment.capture(1)
    aesthetics = {} if x is None else {"x": x}
    if y is not None:
        aesthetics["y"] = y

    def is_mapping(value: Any) -> bool:
        """
        Return True if value is not enclosed in I() function
        """
        with suppress(AttributeError):
            return not (value.startswith("I(") and value.endswith(")"))
        return True

    def I(value: Any) -> Any:
        return value

    I_env = EvalEnvironment([{"I": I}])

    for ae in kwargs.keys() & ALL_AESTHETICS:
        value = kwargs[ae]
        if is_mapping(value):
            aesthetics[ae] = value
        else:
            kwargs[ae] = I_env.eval(value)

    # List of geoms
    if isinstance(geom, str):
        geom = [geom]
    elif isinstance(geom, tuple):
        geom = list(geom)

    if data is None:
        data = pd.DataFrame()

    # Work out plot data, and modify aesthetics, if necessary
    def replace_auto(lst: list[str], str2: str) -> list[str]:
        """
        Replace all occurences of 'auto' in with str2
        """
        for i, value in enumerate(lst):
            if value == "auto":
                lst[i] = str2
        return lst

    if "auto" in geom:
        if "sample" in aesthetics:
            replace_auto(geom, "qq")
        elif y is None:
            # If x is discrete we choose geom_bar &
            # geom_histogram otherwise. But we need to
            # evaluate the mapping to find out the dtype
            env = environment.with_outer_namespace({"factor": pd.Categorical})

            if isinstance(aesthetics["x"], str):
                try:
                    x = env.eval(aesthetics["x"], inner_namespace=data)
                except Exception:
                    msg = "Could not evaluate aesthetic 'x={}'"
                    raise PlotnineError(msg.format(aesthetics["x"]))
            elif not hasattr(aesthetics["x"], "dtype"):
                x = np.asarray(aesthetics["x"])

            if array_kind.discrete(x):
                replace_auto(geom, "bar")
            else:
                replace_auto(geom, "histogram")

        else:
            if x is None:
                if isinstance(aesthetics["y"], typing.Sized):
                    aesthetics["x"] = range(len(aesthetics["y"]))
                    xlab = "range(len(y))"
                    ylab = "y"
                else:
                    # We could solve the issue in layer.compute_asthetics
                    # but it is not worth the extra complexity
                    raise PlotnineError("Cannot infer how long x should be.")
            replace_auto(geom, "point")

    p: Ggplot = ggplot(data, aes(**aesthetics), environment=environment)

    def get_facet_type(facets: str) -> Literal["grid", "wrap", "null"]:
        with suppress(PlotnineError):
            parse_grid_facets(facets)
            return "grid"

        with suppress(PlotnineError):
            parse_wrap_facets(facets)
            return "wrap"

        warn(
            "Could not determine the type of faceting, "
            "therefore no faceting.",
            PlotnineWarning,
        )
        return "null"

    if facets:
        facet_type = get_facet_type(facets)
        if facet_type == "grid":
            p += facet_grid(facets, margins=margins)
        elif facet_type == "wrap":
            p += facet_wrap(facets)
        else:
            p += facet_null()

    # Add geoms
    for g in geom:
        geom_name = f"geom_{g}"
        geom_klass = Registry[geom_name]
        stat_name = "stat_{}".format(geom_klass.DEFAULT_PARAMS["stat"])
        stat_klass = Registry[stat_name]
        # find params
        recognized = kwargs.keys() & (
            geom_klass.DEFAULT_PARAMS.keys()
            | geom_klass.aesthetics()
            | stat_klass.DEFAULT_PARAMS.keys()
            | stat_klass.aesthetics()
        )
        recognized = recognized - aesthetics.keys()
        params = {ae: kwargs[ae] for ae in recognized}
        p += geom_klass(**params)

    # pd.Series objects have name attributes. In a dataframe, the
    # series have the name of the column.
    labels = {}
    for ae in SCALED_AESTHETICS & kwargs.keys():
        with suppress(AttributeError):
            labels[ae] = kwargs[ae].name

    with suppress(AttributeError):
        labels["x"] = xlab if xlab is not None else x.name  # type: ignore

    with suppress(AttributeError):
        labels["y"] = ylab if ylab is not None else y.name  # type: ignore

    if main is not None:
        labels["title"] = main

    if log:
        if "x" in log:
            p += scale_x_log10()

        if "y" in log:
            p += scale_y_log10()

    if labels:
        p += labs(**labels)

    if asp:
        p += theme(aspect_ratio=asp)

    if xlim:
        p += lims(x=xlim)

    if ylim:
        p += lims(y=ylim)

    return p



from __future__ import annotations

import hashlib
import typing
from warnings import warn

import numpy as np
import pandas as pd
from mizani.bounds import rescale

from ..exceptions import PlotnineWarning
from ..mapping.aes import rename_aesthetics
from ..scales.scale import scale_continuous
from .guide import guide

if typing.TYPE_CHECKING:
    from plotnine.typing import ScaleContinuous


class guide_colorbar(guide):
    """
    Guide colorbar

    Parameters
    ----------
    barwidth : float
        Width (in pixels) of the colorbar.
    barheight : float
        Height (in pixels) of the colorbar. The height is multiplied by
        a factor of 5.
    nbin : int
        Number of bins for drawing a colorbar. A larger value yields
        a smoother colorbar. Default is 20.
    raster : bool
        Whether to render the colorbar as a raster object.
    ticks : bool
        Whether tick marks on colorbar should be visible.
    draw_ulim : bool
        Whether to show the upper limit tick marks.
    draw_llim : bool
        Whether to show the lower limit tick marks.
    direction : str in ``['horizontal', 'vertical']``
        Direction of the guide.
    kwargs : dict
        Parameters passed on to :class:`.guide`

    """

    # bar
    barwidth = None
    barheight = None
    nbin = 20  # maximum number of bins
    raster = True

    # ticks
    ticks = True
    draw_ulim = True
    draw_llim = True

    # parameter
    available_aes = {"colour", "color", "fill"}

    def train(self, scale: ScaleContinuous, aesthetic=None):
        if aesthetic is None:
            aesthetic = scale.aesthetics[0]

        # Do nothing if scales are inappropriate
        if set(scale.aesthetics) & self.available_aes == 0:
            warn("colorbar guide needs appropriate scales.", PlotnineWarning)
            return None

        if not isinstance(scale, scale_continuous):
            warn("colorbar guide needs continuous scales", PlotnineWarning)
            return None

        # value = breaks (numeric) is used for determining the
        # position of ticks
        limits = scale.limits
        breaks = scale.get_bounded_breaks()

        if not len(breaks):
            return None

        self.key = pd.DataFrame(
            {
                aesthetic: scale.map(breaks),
                "label": scale.get_labels(breaks),
                "value": breaks,
            }
        )

        bar = np.linspace(limits[0], limits[1], self.nbin)
        self.bar = pd.DataFrame({"color": scale.map(bar), "value": bar})

        labels = " ".join(str(x) for x in self.key["label"])
        info = "\n".join(
            [
                self.title,
                labels,
                " ".join(self.bar["color"].tolist()),
                self.__class__.__name__,
            ]
        )
        self.hash = hashlib.md5(info.encode("utf-8")).hexdigest()
        return self

    def merge(self, other):
        """
        Simply discards the other guide
        """
        return self

    def create_geoms(self, plot):
        """
        Return self if colorbar will be drawn and None if not

        This guide is not geom based
        """
        for l in plot.layers:
            exclude = set()
            if isinstance(l.show_legend, dict):
                l.show_legend = rename_aesthetics(l.show_legend)
                exclude = {ae for ae, val in l.show_legend.items() if not val}
            elif l.show_legend not in (None, True):
                continue

            matched = self.legend_aesthetics(l, plot)

            # layer uses guide
            if set(matched) - exclude:
                break
        # no break, no layer uses this guide
        else:
            return None

        return self

    def draw(self):
        """
        Draw guide

        Returns
        -------
        out : matplotlib.offsetbox.Offsetbox
            A drawing of this legend
        """
        from matplotlib.offsetbox import (
            DrawingArea,
            HPacker,
            TextArea,
            VPacker,
        )

        obverse = slice(0, None)
        reverse = slice(None, None, -1)
        nbars = len(self.bar)
        direction = self.direction
        colors = self.bar["color"].tolist()
        labels = self.key["label"].tolist()
        _targets = self.theme._targets
        _property = self.theme.themeables.property

        # 1.45 makes the default colourbar wider than the
        # legend entry boxes.
        width = (self.barwidth or _property("legend_key_width")) * 1.45
        height = (self.barheight or _property("legend_key_height")) * 1.45

        height *= 5
        length = height

        # When there is more than one guide, we keep
        # record of all of them using lists
        if "legend_title" not in _targets:
            _targets["legend_title"] = []
        if "legend_text_colorbar" not in _targets:
            _targets["legend_text_colorbar"] = []

        # .5 puts the ticks in the middle of the bars when
        # raster=False. So when raster=True the ticks are
        # in between interpolation points and the matching is
        # close though not exactly right.
        _from = self.bar["value"].min(), self.bar["value"].max()
        tick_locations = (
            rescale(self.key["value"], (0.5, nbars - 0.5), _from)
            * length
            / nbars
        )

        if direction == "horizontal":
            width, height = height, width
            length = width

        if self.reverse:
            colors = colors[::-1]
            labels = labels[::-1]
            tick_locations = length - tick_locations[::-1]

        # title #
        title_box = TextArea(self.title, textprops={"color": "black"})
        _targets["legend_title"].append(title_box)

        # colorbar and ticks #
        da = DrawingArea(width, height, 0, 0)
        if self.raster:
            add_interpolated_colorbar(da, colors, direction)
        else:
            add_segmented_colorbar(da, colors, direction)

        if self.ticks:
            _locations = tick_locations
            if not self.draw_ulim:
                _locations = _locations[:-1]

            if not self.draw_llim:
                _locations = _locations[1:]

            add_ticks(da, _locations, direction)

        # labels #
        if self.label:
            labels_da, legend_text = create_labels(
                da, labels, tick_locations, direction
            )
            _targets["legend_text_colorbar"].extend(legend_text)
        else:
            labels_da = DrawingArea(0, 0)

        # colorbar + labels #
        if direction == "vertical":
            packer, align = HPacker, "bottom"
            align = "center"
        else:
            packer, align = VPacker, "right"
            align = "center"
        slc = obverse if self.label_position == "right" else reverse
        if self.label_position in ("right", "bottom"):
            slc = obverse
        else:
            slc = reverse
        main_box = packer(
            children=[da, labels_da][slc],
            sep=self._label_margin,
            align=align,
            pad=0,
        )

        # title + colorbar(with labels) #
        lookup = {
            "right": (HPacker, reverse),
            "left": (HPacker, obverse),
            "bottom": (VPacker, reverse),
            "top": (VPacker, obverse),
        }
        packer, slc = lookup[self.title_position]
        children = [title_box, main_box][slc]
        box = packer(
            children=children,
            sep=self._title_margin,
            align=self._title_align,
            pad=0,
        )
        return box


def add_interpolated_colorbar(da, colors, direction):
    """
    Add 'rastered' colorbar to DrawingArea
    """
    from matplotlib.collections import QuadMesh
    from matplotlib.colors import ListedColormap

    # Special case that arises due to not so useful
    # aesthetic mapping.
    if len(colors) == 1:
        colors = [colors[0], colors[0]]

    # Number of horizontal egdes(breaks) in the grid
    # No need to create more nbreak than colors, provided
    # no. of colors = no. of breaks = no. of cmap colors
    # the shading does a perfect interpolation
    nbreak = len(colors)

    if direction == "vertical":
        mesh_width = 1
        mesh_height = nbreak - 1
        linewidth = da.height / mesh_height
        # Construct rectangular meshgrid
        # The values(Z) at each vertex are just the
        # normalized (onto [0, 1]) vertical distance
        x = np.array([0, da.width])
        y = np.arange(0, nbreak) * linewidth
        X, Y = np.meshgrid(x, y)
        Z = Y / y.max()
    else:
        mesh_width = nbreak - 1
        mesh_height = 1
        linewidth = da.width / mesh_width
        x = np.arange(0, nbreak) * linewidth
        y = np.array([0, da.height])
        X, Y = np.meshgrid(x, y)
        Z = X / x.max()

    # As a 3D (mesh_width x mesh_height x 2) coordinates array
    coordinates = np.stack([X, Y], axis=-1)
    cmap = ListedColormap(colors)
    coll = QuadMesh(
        coordinates,
        antialiased=False,
        shading="gouraud",
        linewidth=0,
        cmap=cmap,
        array=Z.ravel(),
    )
    da.add_artist(coll)


def add_segmented_colorbar(da, colors, direction):
    """
    Add 'non-rastered' colorbar to DrawingArea
    """
    from matplotlib.collections import PolyCollection

    nbreak = len(colors)
    if direction == "vertical":
        linewidth = da.height / nbreak
        verts = []
        x1, x2 = 0, da.width
        for i in range(nbreak):
            y1 = i * linewidth
            y2 = y1 + linewidth
            verts.append(((x1, y1), (x1, y2), (x2, y2), (x2, y1)))
    else:
        linewidth = da.width / nbreak
        verts = []
        y1, y2 = 0, da.height
        for i in range(nbreak):
            x1 = i * linewidth
            x2 = x1 + linewidth
            verts.append(((x1, y1), (x1, y2), (x2, y2), (x2, y1)))

    coll = PolyCollection(
        verts, facecolors=colors, linewidth=0, antialiased=False
    )
    da.add_artist(coll)


def add_ticks(da, locations, direction):
    from matplotlib.collections import LineCollection

    segments = []
    if direction == "vertical":
        x1, x2, x3, x4 = np.array([0.0, 1 / 5, 4 / 5, 1.0]) * da.width
        for y in locations:
            segments.extend(
                [
                    ((x1, y), (x2, y)),
                    ((x3, y), (x4, y)),
                ]
            )
    else:
        y1, y2, y3, y4 = np.array([0.0, 1 / 5, 4 / 5, 1.0]) * da.height
        for x in locations:
            segments.extend(
                [
                    ((x, y1), (x, y2)),
                    ((x, y3), (x, y4)),
                ]
            )

    coll = LineCollection(
        segments, color="#CCCCCC", linewidth=1, antialiased=False
    )
    da.add_artist(coll)


def create_labels(da, labels, locations, direction):
    """
    Return an OffsetBox with label texts
    """
    # The box dimensions are determined by the size of
    # the text objects. We put two dummy children at
    # either end to gaurantee that when center packed
    # the labels in the labels_box matchup with the ticks.
    from matplotlib.text import Text

    from .._mpl.offsetbox import MyAuxTransformBox

    fontsize = 9
    labels_box = MyAuxTransformBox()
    xs, ys = [0] * len(labels), locations
    ha, va = "left", "center"

    x1, y1 = 0, 0
    x2, y2 = 0, da.height
    if direction == "horizontal":
        xs, ys = ys, xs
        ha, va = "center", "top"
        x2, y2 = da.width, 0
    txt1 = Text(x1, y1, "", horizontalalignment=ha, verticalalignment=va)
    txt2 = Text(x2, y2, "", horizontalalignment=ha, verticalalignment=va)
    labels_box.add_artist(txt1)
    labels_box.add_artist(txt2)

    legend_text = []
    for i, (x, y, text) in enumerate(zip(xs, ys, labels)):
        txt = Text(
            x,
            y,
            text,
            size=fontsize,
            horizontalalignment=ha,
            verticalalignment=va,
        )
        labels_box.add_artist(txt)
        legend_text.append(txt)
    return labels_box, legend_text


guide_colourbar = guide_colorbar



from contextlib import suppress

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineError
from .stat import stat
from .stat_density import compute_density, stat_density


@document
class stat_ydensity(stat):
    """
    Density estimate

    {usage}

    Parameters
    ----------
    {common_parameters}
    kernel : str, optional (default: 'gaussian')
        Kernel used for density estimation. One of::

            'biweight'
            'cosine'
            'cosine2'
            'epanechnikov'
            'gaussian'
            'triangular'
            'triweight'
            'uniform'

    adjust : float, optional (default: 1)
        An adjustment factor for the ``bw``. Bandwidth becomes
        :py:`bw * adjust`.
        Adjustment of the bandwidth.
    trim : bool, optional (default: False)
        This parameter only matters if you are displaying multiple
        densities in one plot. If :py:`False`, the default, each
        density is computed on the full range of the data. If
        :py:`True`, each density is computed over the range of that
        group; this typically means the estimated x values will not
        line-up, and hence you won't be able to stack density values.
    n : int, optional(default: 1024)
        Number of equally spaced points at which the density is to
        be estimated. For efficient computation, it should be a power
        of two.
    bw : str or float, optional (default: 'nrd0')
        The bandwidth to use, If a float is given, it is the bandwidth.
        The :py:`str` choices are::

            'normal_reference'
            'scott'
            'silverman'

        ``nrd0`` is a port of ``stats::bw.nrd0`` in R; it is eqiuvalent
        to ``silverman`` when there is more than 1 value in a group.
    scale : (default: area)
        How to scale the violins. The options are::

            'area'   # all violins have the same area, before
                     # trimming the tails.

            'count'  # areas are scaled proportionally to the number
                     # of observations.

            'width'  # all violins have the same maximum width.

    See Also
    --------
    plotnine.geoms.geom_violin
    statsmodels.nonparametric.kde.KDEUnivariate
    statsmodels.nonparametric.kde.KDEUnivariate.fit
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

         'width'        # Maximum width of density, [0, 1] range.
         'violinwidth'  # Shape of the violin

    Calculated aesthetics are accessed using the `after_stat` function.
    e.g. :py:`after_stat('width')`.
    """
    REQUIRED_AES = {"x", "y"}
    NON_MISSING_AES = {"weight"}
    DEFAULT_PARAMS = {
        "geom": "violin",
        "position": "dodge",
        "na_rm": False,
        "adjust": 1,
        "kernel": "gaussian",
        "n": 1024,
        "trim": True,
        "bw": "nrd0",
        "scale": "area",
    }
    DEFAULT_AES = {"weight": None}
    CREATES = {"width", "violinwidth"}

    def setup_params(self, data):
        params = self.params.copy()

        valid_scale = ("area", "count", "width")
        if params["scale"] not in valid_scale:
            msg = "Parameter scale should be one of {}"
            raise PlotnineError(msg.format(valid_scale))

        lookup = {
            "biweight": "biw",
            "cosine": "cos",
            "cosine2": "cos2",
            "epanechnikov": "epa",
            "gaussian": "gau",
            "triangular": "tri",
            "triweight": "triw",
            "uniform": "uni",
        }

        with suppress(KeyError):
            params["kernel"] = lookup[params["kernel"].lower()]

        if params["kernel"] not in lookup.values():
            msg = (
                "kernel should be one of {}. "
                "You may use the abbreviations {}"
            )
            raise PlotnineError(msg.format(lookup.keys(), lookup.values()))

        missing_params = stat_density.DEFAULT_PARAMS.keys() - params.keys()
        for key in missing_params:
            params[key] = stat_density.DEFAULT_PARAMS[key]

        return params

    @classmethod
    def compute_panel(cls, data, scales, **params):
        data = super(cls, cls).compute_panel(data, scales, **params)

        if not len(data):
            return data

        if params["scale"] == "area":
            data["violinwidth"] = data["density"] / data["density"].max()
        elif params["scale"] == "count":
            data["violinwidth"] = (
                data["density"]
                / data["density"].max()
                * data["n"]
                / data["n"].max()
            )
        elif params["scale"] == "width":
            data["violinwidth"] = data["scaled"]
        else:
            msg = "Unknown scale value '{}'"
            raise PlotnineError(msg.format(params["scale"]))

        return data

    @classmethod
    def compute_group(cls, data, scales, **params):
        n = len(data)
        if n == 0:
            return pd.DataFrame()

        weight = data.get("weight")

        if params["trim"]:
            range_y = data["y"].min(), data["y"].max()
        else:
            range_y = scales.y.dimension()

        dens = compute_density(data["y"], weight, range_y, **params)

        if not len(dens):
            return dens

        dens["y"] = dens["x"]
        dens["x"] = np.mean([data["x"].min(), data["x"].max()])

        # Compute width if x has multiple values
        if len(np.unique(data["x"])) > 1:
            dens["width"] = np.ptp(data["x"]) * 0.9

        return dens



import pandas as pd

from plotnine import (
    aes,
    after_stat,
    geom_density_2d,
    geom_point,
    ggplot,
    lims,
    scale_size_radius,
    stat_density_2d,
)

n = 20
adj = n // 4

df = pd.DataFrame({"x": range(n), "y": range(n)})

p0 = ggplot(df, aes("x", "y")) + lims(x=(-adj, n + adj), y=(-adj, n + adj))


def test_contours():
    p = p0 + geom_density_2d(aes(color=after_stat("level")))
    assert p == "contours"


def test_points():
    p = (
        p0
        + geom_point(
            aes(fill=after_stat("density"), size=after_stat("density")),
            stat="density_2d",
            stroke=0,
            n=16,
            contour=False,
        )
        + scale_size_radius(range=(0, 6))
    )

    assert p == "points"


def test_polygon():
    p = p0 + stat_density_2d(aes(fill=after_stat("level")), geom="polygon")
    assert p == "polygon"



from __future__ import annotations

import typing

from ..doctools import document
from .geom_path import geom_path

if typing.TYPE_CHECKING:
    import pandas as pd


@document
class geom_line(geom_path):
    """
    Connected points

    {usage}

    Parameters
    ----------
    {common_parameters}

    See Also
    --------
    plotnine.geoms.geom_path : For documentation of other parameters.
    """

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        return data.sort_values(["PANEL", "group", "x"])



from typing import Any

#: Development flag, e.g. set to ``True`` to prevent
#: the queuing up of figures when errors happen.
close_all_figures = False

#: Theme used when none is added to the ggplot object
current_theme = None

#: The base font family for all text that is part of the theme.
#: Default is sans-serif and one is choosen automatically from
#: rcParams["font.san-serif"]
base_family = "sans-serif"

#: Default aspect ratio used by the themes
aspect_ratio = "auto"

#: Default DPI used by the themes
dpi = 100

#: Default figure size inches
figure_size = (640 / dpi, 480 / dpi)

#: A size that is proportional of the figure width and
#: is used by some themes to determine other margins
base_margin = 0.01


def get_option(name: str) -> Any:
    """
    Get package option

    Parameters
    ----------
    name : str
        Name of the option
    """
    d = globals()

    if name in {"get_option", "set_option"} or name not in d:
        from .exceptions import PlotnineError

        raise PlotnineError(f"Unknown option {name}")

    return d[name]


def set_option(name: str, value: Any) -> Any:
    """
    Set package option

    Parameters
    ----------
    name : str
        Name of the option
    value : object
        New value of the option

    Returns
    -------
    old : object
        Old value of the option
    """
    d = globals()

    if name in {"get_option", "set_option"} or name not in d:
        from .exceptions import PlotnineError

        raise PlotnineError(f"Unknown option {name}")

    old = d[name]
    d[name] = value
    return old



from ..doctools import document
from .geom_path import geom_path


@document
class geom_density_2d(geom_path):
    """
    2D density estimate

    This is a 2d version of :class:`~plotnine.geoms.geom_density`.

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_PARAMS = {
        "stat": "density_2d",
        "position": "identity",
        "na_rm": False,
    }



from ..options import get_option
from .elements import element_rect, element_text
from .theme import theme


class theme_matplotlib(theme):
    """
    The default matplotlib look and feel.

    The theme can be used (and has the same parameter
    to customize) like a :class:`matplotlib.rc_context` manager.

    Parameters
    ----------
    rc :  dict, optional
        rcParams which should be applied on top of
        mathplotlib default.
    fname : str, optional
        Filename to a matplotlibrc file
    use_defaults : bool
        If `True` (the default) resets the plot setting
        to the (current) `matplotlib.rcParams` values
    """

    def __init__(self, rc=None, fname=None, use_defaults=True):
        import matplotlib as mpl

        m = get_option("base_margin")
        base_size = mpl.rcParams.get("font.size", 11)
        super().__init__(
            text=element_text(
                size=base_size,
                linespacing=1,
                rotation=0,
            ),
            aspect_ratio=get_option("aspect_ratio"),
            dpi=get_option("dpi"),
            figure_size=get_option("figure_size"),
            axis_text=element_text(margin={"t": 2.4, "r": 2.4, "units": "pt"}),
            axis_title_x=element_text(
                va="bottom", ha="center", margin={"t": m, "units": "fig"}
            ),
            axis_title_y=element_text(
                angle=90,
                va="center",
                ha="left",
                margin={"r": m, "units": "fig"},
            ),
            legend_box="auto",
            legend_box_just="auto",
            legend_box_margin=0,
            legend_box_spacing=m * 3,
            legend_direction="auto",
            legend_entry_spacing_x=5,
            legend_entry_spacing_y=2,
            legend_key=element_rect(fill="None", colour="None"),
            legend_key_size=16,
            legend_margin=0,
            legend_position="right",
            legend_spacing=10,
            legend_text=element_text(
                margin={"t": 3, "b": 3, "l": 3, "r": 3, "units": "pt"}
            ),
            legend_title=element_text(
                ha="left",
                margin={
                    "t": m,
                    "b": m,
                    "l": m,
                    "r": m,
                    "units": "fig",
                },
            ),
            legend_title_align="auto",
            panel_spacing=m,
            plot_caption=element_text(
                ha="right",
                va="bottom",
                ma="right",
                margin={"t": m, "units": "fig"},
            ),
            plot_margin=m,
            plot_subtitle=element_text(
                size=base_size * 0.9,
                ha="left",
                va="top",
                ma="left",
                margin={"b": m, "units": "fig"},
            ),
            plot_title=element_text(
                ha="left",
                va="top",
                ma="left",
                margin={"b": m, "units": "fig"},
            ),
            strip_align=0,
            strip_background=element_rect(
                fill="#D9D9D9", color="black", size=0.72
            ),
            strip_text=element_text(
                margin={
                    "t": 1 / 3,
                    "b": 1 / 3,
                    "l": 1 / 3,
                    "r": 1 / 3,
                    "units": "lines",
                },
            ),
            strip_text_y=element_text(rotation=-90),
            complete=True,
        )

        if use_defaults:
            _copy = mpl.rcParams.copy()
            if "tk.pythoninspect" in _copy:
                del _copy["tk.pythoninspect"]
            self._rcParams.update(_copy)

        if fname:
            self._rcParams.update(mpl.rc_params_from_file(fname))
        if rc:
            self._rcParams.update(rc)



import numpy as np
import numpy.testing as npt
import pandas as pd

from plotnine import aes, geom_point, ggplot, stat_ellipse
from plotnine.stats.stat_ellipse import cov_trob

df = pd.DataFrame(
    {
        "x": [1, 2, 3, 4, 5],
        "y": [1, 4, 3, 6, 7],
        "z": [3, 4, 3, 2, 6],
    }
)


def test_ellipse():
    p = (
        ggplot(df, aes("x", "y"))
        + geom_point()
        + stat_ellipse(type="t")
        + stat_ellipse(type="norm", color="red")
        + stat_ellipse(type="euclid", color="blue")
    )

    assert p == "ellipse"


def test_cov_trob_2d():
    x = np.array(df[["x", "y"]])
    res = cov_trob(x, cor=True)

    # Values from MASS::cov.trob
    n_obs = 5
    center = [3.110130, 4.359847]
    cov = [[1.979174, 2.812684], [2.812684, 4.584488]]
    cor = [[1.0000000, 0.9337562], [0.9337562, 1.0000000]]

    assert res["n_obs"] == n_obs
    npt.assert_allclose(res["center"], center)
    npt.assert_allclose(res["cov"], cov)
    npt.assert_allclose(res["cor"], cor)


def test_cov_trob_3d():
    x = np.array(df[["x", "y", "z"]])
    res = cov_trob(x, cor=True)

    # Values from MASS::cov.trob
    n_obs = 5
    center = [2.844500, 3.930879, 3.543190]
    cov = [
        [1.9412275, 2.713547, 0.7242778],
        [2.7135469, 4.479363, 1.2210262],
        [0.7242778, 1.221026, 1.6008466],
    ]
    cor = [
        [1.0000000, 0.9202185, 0.4108583],
        [0.9202185, 1.0000000, 0.4559760],
        [0.4108583, 0.4559760, 1.0000000],
    ]

    assert res["n_obs"] == n_obs
    npt.assert_allclose(res["center"], center, rtol=1e-6)
    npt.assert_allclose(res["cov"], cov, rtol=1e-6)
    npt.assert_allclose(res["cor"], cor, rtol=1e-6)



import numpy as np
import pandas as pd

from plotnine import (
    aes,
    after_stat,
    geom_point,
    geom_pointdensity,
    ggplot,
    scale_size_radius,
)

n = 16  # Some even number > 2

df = pd.DataFrame({"x": range(n), "y": np.repeat(range(n // 2), 2)})

p0 = ggplot(df, aes("x", "y"))


def test_pointdensity():
    p = p0 + geom_pointdensity(size=10)
    assert p == "contours"


def test_points():
    p = (
        p0
        + geom_point(
            aes(fill=after_stat("density"), size=after_stat("density")),
            stat="pointdensity",
        )
        + scale_size_radius(range=(10, 20))
    )

    assert p == "points"
